#!/usr/bin/env python
# adopted with few changes from Tim Mooney's mda.py

# version 2.1 Tim Mooney 2/15/2012
# merge of mda.py and mda_f.py
# - supports reading, writing, and arithmetic operations for
#   up to 4-dimensional MDA files.

from __future__ import print_function

__version__ = '2.1'

import sys
import os
import string

import numpy


have_fast_xdr = False
try:
    import f_xdrlib as xdr
    have_fast_xdr = True
except:
    import xdrlib as xdr

try:
    import Tkinter
    have_Tkinter = True
except:
    have_Tkinter = False
#     try:
#         import wx
#         have_wx = True
#     except:
#         have_wx = False

if have_Tkinter:
    import tkFileDialog

# If we can import numpy, and if caller asks us to use it, we'll
# return data in numpy arrays.  Otherwise, we'll return data in lists.

import copy

################################################################################
# classes
# scanDim holds all of the data associated with a single execution of a single sscan record.
class scanDim:
    def __init__(self):
        self.rank = 0            # [1..n]  1 means this is the "innermost" or only scan dimension
        self.dim = 0            # dimensionality of data (numerically same as rank)
        self.npts = 0            # number of data points planned
        self.curr_pt = 0        # number of data points actually acquired
        self.plower_scans = 0    # file offsets of next lower rank scans
        self.name = ""            # name of sscan record that acquired the data
        self.time = ""            # time at which scan (dimension) started
        self.np = 0                # number of positioners
        self.p = []                # list of scanPositioner instances
        self.nd = 0                # number of detectors
        self.d = []                # list of scanDetector instances
        self.nt = 0                # number of detector triggers
        self.t = []                # list of scanTrigger instances

    def __str__(self):
        if self.name != '':
            s = "%dD data from \"%s\" acquired on %s:\n%d/%d pts; %d positioners, %d detectors" % (
                self.dim, self.name, self.time, self.curr_pt, self.npts, self.np, self.nd)
        else:
            s = "%dD data (not read in)" % (self.dim)

        return s

# scanPositioner holds all the information associated with a single positioner, and
# all the data written and acquired by that positioner during an entire (possibly
# multidimensional) scan.
class scanPositioner:
    def __init__(self):
        self.number = 0                # positioner number in sscan record
        self.fieldName = ""            # name of sscanRecord PV
        self.name = ""                # name of EPICS PV this positioner wrote to
        self.desc = ""                # description of 'name' PV
        self.step_mode = ""            # 'LINEAR', 'TABLE', or 'FLY'
        self.unit = ""                # units of 'name' PV
        self.readback_name = ""        # name of EPICS PV this positioner read from, if any
        self.readback_desc = ""        # description of 'readback_name' PV
        self.readback_unit = ""        # units of 'readback_name' PV
        self.data = []                # list of values written to 'name' PV.  If rank==2, lists of lists, etc.

    def __str__(self):
        data = self.data

        n = data.ndim
        if n==1:
            dimString = '(' + str(data.shape[0]) + ')'
        else:
            dimString = str(data.shape)

        s = "positioner <scanRecord>.%s\nPV name  '%s'\nPV desc. '%s'\nPV units    '%s'\nstep mode: %s\nRB name  '%s'\nRB desc. '%s'\nRB units    '%s'\ndata: %dD array %s\n" % (self.fieldName,
        self.name, self.desc, self.unit, self.step_mode, self.name, self.desc, self.unit, n, dimString)
        return s

# scanDetector holds all the information associated with a single detector, and
# all the data acquired by that detector during an entire (possibly multidimensional) scan.
class scanDetector:
    def __init__(self):
        self.number = 0            # detector number in sscan record
        self.fieldName = ""        # name of sscanRecord PV
        self.name = ""            # name of EPICS PV this detector read from
        self.desc = ""            # description of 'name' PV
        self.unit = ""            # units of 'name' PV
        self.data = []            # list of values read from 'name' PV.  If rank==2, lists of lists, etc.

    def __str__(self):
        data = self.data
        n = data.ndim
        if n==1:
            dimString = '(' + str(data.shape[0]) + ')'
        else:
            dimString = str(data.shape)
        s = "detector <scanRecord>.%s\nPV name  '%s'\nPV desc. '%s'\nPV units    '%s'\ndata: %dD array %s\n" % (self.fieldName,
        self.name, self.desc, self.unit, n, dimString)
        return s

# scanTrigger holds all the information associated with a single detector trigger.
class scanTrigger:
    def __init__(self):
        self.number = 0            # detector-trigger number in sscan record
        self.name = ""            # name of sscanRecord PV
        self.command = 0.0        # value written to 'name' PV

    def __str__(self):
        s = "trigger %d (%s), command=%f\n" % (self.number,
            self.name, self.command)
        return s

# scanBuf is a private data structure used to assemble data that will be written to an MDA file.
class scanBuf:
    def __init__(self):
        self.npts = 0
        self.offset = 0
        self.bufLen = 0
        self.preamble = None
        self.pLowerScans = []
        self.pLowerScansBuf = ""
        self.postamble = None
        self.data = None
        self.inner = []    # inner scans, if any

# mdaBuf is a private data structure used to assemble data that will be written to an MDA file.
class mdaBuf:
    def __init__(self):
        self.header = None
        self.pExtra = None    # file offset to extraPV section
        self.scan = None
        self.extraPV = None    # extraPV section

################################################################################
# read MDA file

# Given a detector number, return the name of the associated sscanRecord PV, 'D01'-'D99'.
# (Currently, only 70 detectors are ever used.)
def detName(i):
    if i < 100:
        return "D%02d"%(i+1)
    else:
        return "?"

# Given a detector number, return the name of the associated sscanRecord PV, for the
# old sscanRecord, which had only 15 detectors 'D1'-'DF'.
def oldDetName(i):
    if i < 15:
        return string.upper("D%s"%(hex(i+1)[2]))
    elif i < 85:
        return "D%02d"%(i-14)
    else:
        return "?"

# Given a positioner number, , return the name of the associated sscanRecord PV, "P1'-'P4'
def posName(i):
    if i < 4:
        return "P%d" % (i+1)
    else:
        return "?"

def verboseData(data, out=sys.stdout):
    if ((len(data)>0) and (type(data[0]) == type([]))):
        for i in len(data):
            verboseData(data[i], out)
    else:
        out.write("[")
        for datum in data:
            if (type(datum) == type(0)):
                out.write(" %d" % datum)
            else:
                out.write(" %.5f" % datum)
        out.write(" ]\n")

def readScan(scanFile, verbose=0, out=sys.stdout, unpacker=None):
    """usage: (scan,num) = readScan(scanFile, verbose=0, out=sys.stdout)"""

    scan = scanDim()    # data structure to hold scan info and data
    buf = scanFile.read(10000) # enough to read scan header
    if unpacker == None:
        u = xdr.Unpacker(buf)
    else:
        u = unpacker
        u.reset(buf)

    scan.rank = u.unpack_int()
    if (scan.rank > 20) or (scan.rank < 0):
        print("* * * readScan('%s'): rank > 20.  probably a corrupt file" % scanFile.name)
        return None

    scan.npts = u.unpack_int()
    scan.curr_pt = u.unpack_int()
    if verbose:
        print("scan.rank = ", scan.rank)
        print("scan.npts = ", scan.npts)
        print("scan.curr_pt = ", scan.curr_pt)

    if (scan.rank > 1):
        # if curr_pt < npts, plower_scans will have garbage for pointers to
        # scans that were planned for but not written
        if have_fast_xdr:
            scan.plower_scans = u.unpack_farray_int(scan.npts)
        else:
            scan.plower_scans = u.unpack_farray(scan.npts, u.unpack_int)
        if verbose:
                        print("scan.plower_scans = ", scan.plower_scans)
    namelength = u.unpack_int()
    scan.name = u.unpack_string()
    if verbose:
                print("scan.name = ", scan.name)
    timelength = u.unpack_int()
    scan.time = u.unpack_string()
    if verbose:
                print("scan.time = ", scan.time)
    scan.np = u.unpack_int()
    if verbose:
                print("scan.np = ", scan.np)
    scan.nd = u.unpack_int()
    if verbose:
                print("scan.nd = ", scan.nd)
    scan.nt = u.unpack_int()
    if verbose:
                print("scan.nt = ", scan.nt)
    for j in range(scan.np):
        scan.p.append(scanPositioner())
        scan.p[j].number = u.unpack_int()
        scan.p[j].fieldName = posName(scan.p[j].number)
        if verbose:
                        print("positioner ", j)
        length = u.unpack_int() # length of name string
        if length:
                        scan.p[j].name = u.unpack_string()
        if verbose:
                        print("scan.p[%d].name = %s" % (j, scan.p[j].name))
        length = u.unpack_int() # length of desc string
        if length:
                        scan.p[j].desc = u.unpack_string()
        if verbose:
                        print("scan.p[%d].desc = %s" % (j, scan.p[j].desc))
        length = u.unpack_int() # length of step_mode string
        if length:
                        scan.p[j].step_mode = u.unpack_string()
        if verbose:
                        print("scan.p[%d].step_mode = %s" % (j, scan.p[j].step_mode))
        length = u.unpack_int() # length of unit string
        if length:
                        scan.p[j].unit = u.unpack_string()
        if verbose:
                        print("scan.p[%d].unit = %s" % (j, scan.p[j].unit))
        length = u.unpack_int() # length of readback_name string
        if length:
                        scan.p[j].readback_name = u.unpack_string()
        if verbose:
                        print("scan.p[%d].readback_name = %s" % (j, scan.p[j].readback_name))
        length = u.unpack_int() # length of readback_desc string
        if length:
                        scan.p[j].readback_desc = u.unpack_string()
        if verbose:
                        print("scan.p[%d].readback_desc = %s" % (j, scan.p[j].readback_desc))
        length = u.unpack_int() # length of readback_unit string
        if length:
                        scan.p[j].readback_unit = u.unpack_string()
        if verbose:
                        print("scan.p[%d].readback_unit = %s" % (j, scan.p[j].readback_unit))

    file_loc_det = scanFile.tell() - (len(buf) - u.get_position())

    for j in range(scan.nd):
        scan.d.append(scanDetector())
        scan.d[j].number = u.unpack_int()
        scan.d[j].fieldName = detName(scan.d[j].number)
        if verbose:
                        print("detector ", j)
        length = u.unpack_int() # length of name string
        if length:
                        scan.d[j].name = u.unpack_string()
        if verbose:
                        print("scan.d[%d].name = %s" % (j, scan.d[j].name))
        length = u.unpack_int() # length of desc string
        if length:
                        scan.d[j].desc = u.unpack_string()
        if verbose:
                        print("scan.d[%d].desc = %s" % (j, scan.d[j].desc))
        length = u.unpack_int() # length of unit string
        if length:
                        scan.d[j].unit = u.unpack_string()
        if verbose:
                        print("scan.d[%d].unit = %s" % (j, scan.d[j].unit))

    for j in range(scan.nt):
        scan.t.append(scanTrigger())
        scan.t[j].number = u.unpack_int()
        if verbose:
                        print("trigger ", j)
        length = u.unpack_int() # length of name string
        if length:
                        scan.t[j].name = u.unpack_string()
        if verbose:
                        print("scan.t[%d].name = %s" % (j, scan.t[j].name))
        scan.t[j].command = u.unpack_float()
        if verbose:
                        print("scan.t[%d].command = %s" % (j, scan.t[j].command))

    ### read data
    # positioners
    file_loc_data = scanFile.tell() - (len(buf) - u.get_position())
    scanFile.seek(file_loc_data)
    buf = scanFile.read(scan.npts * (scan.np * 8 + scan.nd *4))
    u.reset(buf)

    if have_fast_xdr:
        data = u.unpack_farray_double(scan.npts*scan.np)
    else:
        data = u.unpack_farray(scan.npts*scan.np, u.unpack_double)
    start = 0
    end = scan.npts
    for j in range(scan.np):
        start = j*scan.npts
        scan.p[j].data = data[start:end]
        start = end
        end += scan.npts

    # detectors
    if have_fast_xdr:
        data = u.unpack_farray_float(scan.npts*scan.nd)
    else:
        data = u.unpack_farray(scan.npts*scan.nd, u.unpack_float)
    start = 0
    end = scan.npts
    for j in range(scan.nd):
        scan.d[j].data = data[start:end]
        start = end
        end += scan.npts

    return (scan, (file_loc_data-file_loc_det))

useDetToDatOffset = 1
def readScanQuick(scanFile, unpacker=None, detToDat_offset=None):
    """usage: readScanQuick(scanFile, unpacker=None)"""

    scan = scanDim()    # data structure to hold scan info and data
    buf = scanFile.read(10000) # enough to read scan header
    if unpacker == None:
        u = xdr.Unpacker(buf)
    else:
        u = unpacker
        u.reset(buf)

    scan.rank = u.unpack_int()
    if (scan.rank > 20) or (scan.rank < 0):
        print("* * * readScanQuick('%s'): rank > 20.  probably a corrupt file" % scanFile.name)
        return None

    scan.npts = u.unpack_int()
    scan.curr_pt = u.unpack_int()

    if (scan.rank > 1):
        if have_fast_xdr:
            scan.plower_scans = u.unpack_farray_int(scan.npts)
        else:
            scan.plower_scans = u.unpack_farray(scan.npts, u.unpack_int)

    namelength = u.unpack_int()
    scan.name = u.unpack_string()
    timelength = u.unpack_int()
    scan.time = u.unpack_string()

    scan.np = u.unpack_int()
    scan.nd = u.unpack_int()
    scan.nt = u.unpack_int()

    for j in range(scan.np):
        scan.p.append(scanPositioner())
        scan.p[j].number = u.unpack_int()
        n = u.unpack_int() # length of name string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of desc string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of step_mode string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of unit string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of readback_name string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of readback_desc string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)
        n = u.unpack_int() # length of readback_unit string
        if n: u.set_position(u.get_position()+4+(n+3)//4*4)

    file_loc_det = scanFile.tell() - (len(buf) - u.get_position())

    if (detToDat_offset == None) or (not useDetToDatOffset):
        for j in range(scan.nd):
            scan.d.append(scanDetector())
            scan.d[j].number = u.unpack_int()
            scan.d[j].fieldName = detName(scan.d[j].number)
            n = u.unpack_int() # length of name string
            if n: u.set_position(u.get_position()+4+(n+3)//4*4)
            n = u.unpack_int() # length of desc string
            if n: u.set_position(u.get_position()+4+(n+3)//4*4)
            n = u.unpack_int() # length of unit string
            if n: u.set_position(u.get_position()+4+(n+3)//4*4)

        for j in range(scan.nt):
            scan.t.append(scanTrigger())
            scan.t[j].number = u.unpack_int()
            n = u.unpack_int() # length of name string
            if n: u.set_position(u.get_position()+4+(n+3)//4*4)
            scan.t[j].command = u.unpack_float()

        ### read data
        # positioners

        file_loc = scanFile.tell() - (len(buf) - u.get_position())
        diff = file_loc - (file_loc_det + detToDat_offset)
        if diff != 0:
            print("oldSeek, newSeek, o-n=", file_loc, file_loc_det + detToDat_offset, diff)
        scanFile.seek(file_loc)
    else:
        for j in range(scan.nd):
            scan.d.append(scanDetector())
        scanFile.seek(file_loc_det + detToDat_offset)

    buf = scanFile.read(scan.npts * (scan.np * 8 + scan.nd *4))
    u.reset(buf)

    if have_fast_xdr:
        data = u.unpack_farray_double(scan.npts*scan.np)
    else:
        data = u.unpack_farray(scan.npts*scan.np, u.unpack_double)
    for j in range(scan.np):
        start = j*scan.npts
        scan.p[j].data = data[j*scan.npts : (j+1)*scan.npts]

    # detectors
    if have_fast_xdr:
        data = u.unpack_farray_float(scan.npts*scan.nd)
    else:
        data = u.unpack_farray(scan.npts*scan.nd, u.unpack_float)
    for j in range(scan.nd):
        scan.d[j].data = data[j*scan.npts : (j+1)*scan.npts]

    return scan

EPICS_types_dict = {
0: "DBR_STRING",
1: "DBR_SHORT",
2: "DBR_FLOAT",
3: "DBR_ENUM",
4: "DBR_CHAR",
5: "DBR_LONG",
6: "DBR_DOUBLE",
7: "DBR_STS_STRING",
8: "DBR_STS_SHORT",
9: "DBR_STS_FLOAT",
10: "DBR_STS_ENUM",
11: "DBR_STS_CHAR",
12: "DBR_STS_LONG",
13: "DBR_STS_DOUBLE",
14: "DBR_TIME_STRING",
15: "DBR_TIME_SHORT",
16: "DBR_TIME_FLOAT",
17: "DBR_TIME_ENUM",
18: "DBR_TIME_CHAR",
19: "DBR_TIME_LONG",
20: "DBR_TIME_DOUBLE",
21: "DBR_GR_STRING",
22: "DBR_GR_SHORT",
23: "DBR_GR_FLOAT",
24: "DBR_GR_ENUM",
25: "DBR_GR_CHAR",
26: "DBR_GR_LONG",
27: "DBR_GR_DOUBLE",
28: "DBR_CTRL_STRING",
29: "DBR_CTRL_SHORT",
30: "DBR_CTRL_FLOAT",
31: "DBR_CTRL_ENUM",
32: "DBR_CTRL_CHAR",
33: "DBR_CTRL_LONG",
34: "DBR_CTRL_DOUBLE"
}

def EPICS_types(n):
    if EPICS_types_dict.has_key(n):
        return EPICS_types_dict[n]
    else:
        return ("Unexpected type %d" % n)

def readMDA(fname=None, maxdim=4, verbose=0, showHelp=0, outFile=None, useNumpy=None, readQuick=False):
    """usage readMDA(fname=None, maxdim=4, verbose=0, showHelp=0, outFile=None, readQuick=False)"""


    dim = []
    if fname is None:
        if have_Tkinter:
            fname = tkFileDialog.Open().show()
        else:
            print("No file specified, and no file dialog could be opened")
            return None
    if (not os.path.isfile(fname)):
        if (not fname.endswith('.mda')):
            fname = fname + '.mda'
        if (not os.path.isfile(fname)):
            print(fname, "not found")
            return None

    if (outFile == None):
        out = sys.stdout
    else:
        out = open(outFile, 'w')

    scanFile = open(fname, 'rb')
    if verbose: out.write("verbose=%d output for MDA file '%s'\n" % (verbose, fname))
    buf = scanFile.read(100)        # to read header for scan of up to 5 dimensions
    u = xdr.Unpacker(buf)

    # read file header
    version = u.unpack_float()
    if verbose: out.write("MDA version = %.3f\n" % version)
    if abs(version - 1.3) > .01:
        print("I can't read MDA version %f.  Is this really an MDA file?" % version)
        return None

    scan_number = u.unpack_int()
    if verbose: out.write("scan_number = %d\n" % scan_number)
    rank = u.unpack_int()
    if verbose: out.write("rank = %d\n" % rank)
    dimensions = u.unpack_farray(rank, u.unpack_int)
    if verbose:
        out.write("dimensions = ")
        verboseData(dimensions, out)
    isRegular = u.unpack_int()
    if verbose: out.write("isRegular = %d\n" % isRegular)
    pExtra = u.unpack_int()
    if verbose: out.write("pExtra = %d (0x%x)\n" % (pExtra, pExtra))
    pmain_scan = scanFile.tell() - (len(buf) - u.get_position())

    # collect 1D data
    scanFile.seek(pmain_scan)
    (s,n) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
    dim.append(s)
    dim[0].dim = 1

    for p in dim[0].p:
        p.data = numpy.array(p.data)
    for d in dim[0].d:
        d.data = numpy.array(d.data)

    if ((rank > 1) and (maxdim > 1)):
        # collect 2D data
        for i in range(dim[0].curr_pt):
            scanFile.seek(dim[0].plower_scans[i])
            if (i==0):
                (s,detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
                dim.append(s)
                dim[1].dim = 2
                # replace data arrays [1,2,3] with [[1,2,3]]
                for j in range(dim[1].np):
                    dim[1].p[j].data = [dim[1].p[j].data]
                for j in range(dim[1].nd):
                    dim[1].d[j].data = [dim[1].d[j].data]
            else:
                if readQuick:
                    s = readScanQuick(scanFile, unpacker=u, detToDat_offset=detToDat)
                else:
                    (s,junk) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
                # append data arrays
                # [ [1,2,3], [2,3,4] ] -> [ [1,2,3], [2,3,4], [3,4,5] ]
                numP = min(s.np, len(dim[1].p))
                if (s.np > numP):
                    print("First scan had %d positioners; This one only has %d." % (s.np, numP))
                for j in range(numP):
                                        dim[1].p[j].data.append(s.p[j].data)
                numD = min(s.nd, len(dim[1].d))
                if (s.nd > numD):
                    print("First scan had %d detectors; This one only has %d." % (s.nd, numD))
                for j in range(numD): dim[1].d[j].data.append(s.d[j].data)

        for p in dim[1].p:
            p.data = numpy.array(p.data)
        for d in dim[1].d:
            d.data = numpy.array(d.data)

    if ((rank > 2) and (maxdim > 2)):
        # collect 3D data
        #print("dim[0].curr_pt=",dim[0].curr_pt)
        for i in range(dim[0].curr_pt):
            #print("i=%d of %d points" % (i, dim[0].curr_pt))
            scanFile.seek(dim[0].plower_scans[i])
            (s1,detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
            #print("s1.curr_pt=", s1.curr_pt)
            for j in range(s1.curr_pt):
                #print("j=%d of %d points" % (j, s1.curr_pt))
                scanFile.seek(s1.plower_scans[j])
                if (j==0) or not readQuick:
                    (s, detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
                else:
                    s = readScanQuick(scanFile, unpacker=u, detToDat_offset=detToDat)
                if ((i == 0) and (j == 0)):
                    dim.append(s)
                    dim[2].dim = 3
                    # replace data arrays [1,2,3] with [[[1,2,3]]]
                    for k in range(dim[2].np):
                        dim[2].p[k].data = [[dim[2].p[k].data]]
                    for k in range(dim[2].nd):
                        dim[2].d[k].data = [[dim[2].d[k].data]]
                else:
                    # append data arrays
                    numP = min(s.np, len(dim[2].p))
                    if (s.np > numP):
                        print("First scan had %d positioners; This one only has %d." % (s.np, numP))
                    for k in range(numP):
                        if j==0: dim[2].p[k].data.append([])
                        dim[2].p[k].data[i].append(s.p[k].data)
                    numD = min(s.nd, len(dim[2].d))
                    if (s.nd > numD):
                        print("First scan had %d detectors; This one only has %d." % (s.nd, numD))
                    for k in range(numD):
                        if j==0: dim[2].d[k].data.append([])
                        dim[2].d[k].data[i].append(s.d[k].data)

        for p in dim[2].p:
            p.data = numpy.array(p.data)
        for d in dim[2].d:
            d.data = numpy.array(d.data)

    if ((rank > 3) and (maxdim > 3)):
        # collect 4D data
        for i in range(dim[0].curr_pt):
            scanFile.seek(dim[0].plower_scans[i])
            (s1, detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
            for j in range(s1.curr_pt):
                scanFile.seek(s1.plower_scans[j])
                (s2, detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
                for k in range(s2.curr_pt):
                    scanFile.seek(s2.plower_scans[k])
                    if (k==0) or not readQuick:
                        (s, detToDat) = readScan(scanFile, max(0,verbose-1), out, unpacker=u)
                    else:
                        s = readScanQuick(scanFile, unpacker=u, detToDat_offset=detToDat)
                    if ((i == 0) and (j == 0) and (k == 0)):
                        dim.append(s)
                        dim[3].dim = 4
                        for m in range(dim[3].np):
                            dim[3].p[m].data = [[[dim[3].p[m].data]]]
                        for m in range(dim[3].nd):
                            dim[3].d[m].data = [[[dim[3].d[m].data]]]
                    else:
                        # append data arrays
                        if j==0 and k==0:
                            for m in range(dim[3].np):
                                dim[3].p[m].data.append([[]])
                                dim[3].p[m].data[i][0].append(s.p[m].data)
                            for m in range(dim[3].nd):
                                dim[3].d[m].data.append([[]])
                                dim[3].d[m].data[i][0].append(s.d[m].data)
                        else:
                            for m in range(dim[3].np):
                                if k==0: dim[3].p[m].data[i].append([])
                                dim[3].p[m].data[i][j].append(s.p[m].data)
                            for m in range(dim[3].nd):
                                if k==0: dim[3].d[m].data[i].append([])
                                dim[3].d[m].data[i][j].append(s.d[m].data)

        for p in dim[3].p:
            p.data = numpy.array(p.data)
        for d in dim[3].d:
            d.data = numpy.array(d.data)



    # Collect scan-environment variables into a dictionary
    dict = {}
    dict['sampleEntry'] = ("description", "unit string", "value", "EPICS_type", "count")
    dict['filename'] = fname
    dict['version'] = version
    dict['scan_number'] = scan_number
    dict['rank'] = rank
    dict['dimensions'] = dimensions
    acq_dimensions = []
    for d in dim:
        acq_dimensions.append(d.curr_pt)
    dict['acquired_dimensions'] = acq_dimensions
    dict['isRegular'] = isRegular
    dict['ourKeys'] = ['sampleEntry', 'filename', 'version', 'scan_number', 'rank', 'dimensions', 'acquired_dimensions', 'isRegular', 'ourKeys']
    if pExtra:
        scanFile.seek(pExtra)
        buf = scanFile.read()       # Read all scan-environment data
        u.reset(buf)
        numExtra = u.unpack_int()
        if verbose: out.write("\nnumber of 'Extra' PV's = %d\n" % numExtra)
        for i in range(numExtra):
            if verbose: out.write("env PV #%d -------\n" % (i))
            name = ''
            n = u.unpack_int()      # length of name string
            if n: name = u.unpack_string()
            if verbose: out.write("\tname = '%s'\n" % name)
            desc = ''
            n = u.unpack_int()      # length of desc string
            if n: desc = u.unpack_string()
            if verbose: out.write("\tdesc = '%s'\n" % desc)
            EPICS_type = u.unpack_int()
            if verbose: out.write("\tEPICS_type = %d (%s)\n" % (EPICS_type, EPICS_types(EPICS_type)))

            unit = ''
            value = ''
            count = 0
            if EPICS_type != 0:   # not DBR_STRING; array is permitted
                count = u.unpack_int()  #
                if verbose: out.write("\tcount = %d\n" % count)
                n = u.unpack_int()      # length of unit string
                if n: unit = u.unpack_string()
                if verbose: out.write("\tunit = '%s'\n" % unit)

            if EPICS_type == 0: # DBR_STRING
                n = u.unpack_int()      # length of value string
                if n: value = u.unpack_string()
            elif EPICS_type == 32: # DBR_CTRL_CHAR
                #value = u.unpack_fstring(count)
                vect = u.unpack_farray(count, u.unpack_int)
                value = ""
                for i in range(len(vect)):
                    # treat the byte array as a null-terminated string
                    if vect[i] == 0: break
                    value = value + chr(vect[i])
            elif EPICS_type == 29: # DBR_CTRL_SHORT
                value = u.unpack_farray(count, u.unpack_int)
            elif EPICS_type == 33: # DBR_CTRL_LONG
                value = u.unpack_farray(count, u.unpack_int)
            elif EPICS_type == 30: # DBR_CTRL_FLOAT
                value = u.unpack_farray(count, u.unpack_float)
            elif EPICS_type == 34: # DBR_CTRL_DOUBLE
                value = u.unpack_farray(count, u.unpack_double)
            if verbose:
                if (EPICS_type == 0):
                    out.write("\tvalue = '%s'\n" % (value))
                else:
                    out.write("\tvalue = ")
                    verboseData(value, out)

            dict[name] = (desc, unit, value, EPICS_type, count)
    scanFile.close()

    dim.reverse()
    dim.append(dict)
    dim.reverse()
    if verbose or showHelp:
        print("\n%s is a %d-D file; %d dimensions read in." % (fname, dim[0]['rank'], len(dim)-1))
        print("dim[0] = dictionary of %d scan-environment PVs" % (len(dim[0])))
        print("   usage: dim[0]['sampleEntry'] ->", dim[0]['sampleEntry'])
        for i in range(1,len(dim)):
            print("dim[%d] = %s" % (i, str(dim[i])))
        print("   usage: dim[1].p[2].data -> 1D array of positioner 2 data")
        print("   usage: dim[2].d[7].data -> 2D array of detector 7 data")

    if showHelp:
        print(" ")
        print("   each scan dimension (i.e., dim[1], dim[2], ...) has the following fields: ")
        print("      time      - date & time at which scan was started: %s" % (dim[1].time))
        print("      name - name of scan record that acquired this dimension: '%s'" % (dim[1].name))
        print("      curr_pt   - number of data points actually acquired: %d" % (dim[1].curr_pt))
        print("      npts      - number of data points requested: %d" % (dim[1].npts))
        print("      nd        - number of detectors for this scan dimension: %d" % (dim[1].nd))
        print("      d[]       - list of detector-data structures")
        print("      np        - number of positioners for this scan dimension: %d" % (dim[1].np))
        print("      p[]       - list of positioner-data structures")
        print("      nt        - number of detector triggers for this scan dimension: %d" % (dim[1].nt))
        print("      t[]       - list of trigger-info structures")

        print(" ")
        print("   each detector-data structure (e.g., dim[1].d[0]) has the following fields: ")
        print("      desc      - description of this detector")
        print("      data      - data list")
        print("      unit      - engineering units associated with this detector")
        print("      fieldName - scan-record field (e.g., 'D01')")

        print(" ")
        print("   each positioner-data structure (e.g., dim[1].p[0]) has the following fields: ")
        print("      desc          - description of this positioner")
        print("      data          - data list")
        print("      step_mode     - scan mode (e.g., Linear, Table, On-The-Fly)")
        print("      unit          - engineering units associated with this positioner")
        print("      fieldName     - scan-record field (e.g., 'P1')")
        print("      name          - name of EPICS PV (e.g., 'xxx:m1.VAL')")
        print("      readback_desc - description of this positioner")
        print("      readback_unit - engineering units associated with this positioner")
        print("      readback_name - name of EPICS PV (e.g., 'xxx:m1.VAL')")

    if (outFile):
        out.close()
    return dim

################################################################################
# skim MDA file to get dimensions (planned and actually acquired), and other info
def skimScan(dataFile):
    """usage: skimScan(dataFile)"""
    scan = scanDim()    # data structure to hold scan info and data
    buf = dataFile.read(10000) # enough to read scan header
    u = xdr.Unpacker(buf)
    scan.rank = u.unpack_int()
    if (scan.rank > 20) or (scan.rank < 0):
        print("* * * skimScan('%s'): rank > 20.  probably a corrupt file" % dataFile.name)
        return None
    scan.npts = u.unpack_int()
    scan.curr_pt = u.unpack_int()
    if (scan.curr_pt == 0):
        #print("mda:skimScan: curr_pt = 0")
        return None
    if (scan.rank > 1):
        if have_fast_xdr:
            scan.plower_scans = u.unpack_farray_int(scan.npts)
        else:
            scan.plower_scans = u.unpack_farray(scan.npts, u.unpack_int)
    namelength = u.unpack_int()
    scan.name = u.unpack_string()
    timelength = u.unpack_int()
    scan.time = u.unpack_string()
    scan.np = u.unpack_int()
    scan.nd = u.unpack_int()
    scan.nt = u.unpack_int()
    return scan

def skimMDA(fname=None, verbose=False):
    """usage skimMDA(fname=None)"""
    #print("skimMDA: filename=", fname)
    dim = []
    if (fname == None):
        print("No file specified")
        return None
    if (not os.path.isfile(fname)):
        if (not fname.endswith('.mda')):
            fname = fname + '.mda'
        if (not os.path.isfile(fname)):
            print(fname, "not found")
            return None

    try:
        dataFile = open(fname, 'rb')
    except:
        print("mda_f:skimMDA: failed to open file '%s'" % fname)
        return None

    buf = dataFile.read(100)        # to read header for scan of up to 5 dimensions
    u = xdr.Unpacker(buf)

    # read file header
    version = u.unpack_float()
#    if version < 1.299 or version > 1.301:
#        print(fname, " has file version", version)
#        return None
    scan_number = u.unpack_int()
    rank = u.unpack_int()
    dimensions = u.unpack_farray(rank, u.unpack_int)
    isRegular = u.unpack_int()
    pExtra = u.unpack_int()
    pmain_scan = dataFile.tell() - (len(buf) - u.get_position())

    # collect 1D data
    dataFile.seek(pmain_scan)
    scan = skimScan(dataFile)
    if (scan == None):
        if verbose: print(fname, "contains no data")
        return None

    dim.append(scan)
    dim[0].dim = 1

    if (rank > 1):
        dataFile.seek(dim[0].plower_scans[0])
        dim.append(skimScan(dataFile))
        if (dim[1]):
            dim[1].dim = 2
        else:
            if verbose: print("had a problem reading 2d from ", fname)
            return None

    if (rank > 2):
        dataFile.seek(dim[1].plower_scans[0])
        dim.append(skimScan(dataFile))
        if (dim[2]):
            dim[2].dim = 3
        else:
            if verbose: print("had a problem reading 3d from ", fname)
            return None

    if (rank > 3):
        dataFile.seek(dim[2].plower_scans[0])
        dim.append(skimScan(dataFile))
        if (dim[3]):
            dim[3].dim = 4
        else:
            if verbose: print("had a problem reading 4d from ", fname)
            return None

    dataFile.close()
    dict = {}
    dict['filename'] = fname
    dict['version'] = version
    dict['scan_number'] = scan_number
    dict['rank'] = rank
    dict['dimensions'] = dimensions
    dimensions = []
    for d in dim:
        dimensions.append(d.curr_pt)
    dict['acquired_dimensions'] = dimensions
    dict['isRegular'] = isRegular
    dim.reverse()
    dim.append(dict)
    dim.reverse()
    return dim

################################################################################
# Write MDA file
def packScanHead(scan):
    s = scanBuf()
    s.npts = scan.npts

    # preamble
    p = xdr.Packer()
    p.pack_int(scan.rank)
    p.pack_int(scan.npts)
    p.pack_int(scan.curr_pt)
    s.preamble = p.get_buffer()

    # file offsets to lower level scans (if any)
    p.reset()
    if (scan.rank > 1):
        # Pack zeros for now, so we'll know how much
        # space the real offsets will use.
        for j in range(scan.npts):
            p.pack_int(0)
    s.pLowerScansBuf = p.get_buffer()

    # postamble
    p.reset()
    n = len(scan.name); p.pack_int(n)
    if (n): p.pack_string(scan.name)
    n = len(scan.time); p.pack_int(n)
    if (n): p.pack_string(scan.time)
    p.pack_int(scan.np)
    p.pack_int(scan.nd)
    p.pack_int(scan.nt)

    for j in range(scan.np):
        p.pack_int(scan.p[j].number)

        n = len(scan.p[j].name); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].name)

        n = len(scan.p[j].desc); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].desc)

        n = len(scan.p[j].step_mode); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].step_mode)

        n = len(scan.p[j].unit); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].unit)

        n = len(scan.p[j].readback_name); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].readback_name)

        n = len(scan.p[j].readback_desc); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].readback_desc)

        n = len(scan.p[j].readback_unit); p.pack_int(n)
        if (n): p.pack_string(scan.p[j].readback_unit)

    for j in range(scan.nd):
        p.pack_int(scan.d[j].number)
        n = len(scan.d[j].name); p.pack_int(n)
        if (n): p.pack_string(scan.d[j].name)
        n = len(scan.d[j].desc); p.pack_int(n)
        if (n): p.pack_string(scan.d[j].desc)
        n = len(scan.d[j].unit); p.pack_int(n)
        if (n): p.pack_string(scan.d[j].unit)

    for j in range(scan.nt):
        p.pack_int(scan.t[j].number)
        n = len(scan.t[j].name); p.pack_int(n)
        if (n): p.pack_string(scan.t[j].name)
        p.pack_float(scan.t[j].command)

    s.postamble = p.get_buffer()
    s.bufLen = len(s.preamble) + len(s.pLowerScansBuf) + len(s.postamble)
    return s

def packScanData(scan, cpt):
    p = xdr.Packer()
    if (len(cpt) == 0): # 1D array
        for i in range(scan.np):
            p.pack_farray(scan.npts, scan.p[i].data, p.pack_double)
        for i in range(scan.nd):
            p.pack_farray(scan.npts, scan.d[i].data, p.pack_float)

    elif (len(cpt) == 1): # 2D array
        j = cpt[0]
        for i in range(scan.np):
            p.pack_farray(scan.npts, scan.p[i].data[j], p.pack_double)
        for i in range(scan.nd):
            p.pack_farray(scan.npts, scan.d[i].data[j], p.pack_float)

    elif (len(cpt) == 2): # 3D array
        j = cpt[0]
        k = cpt[1]
        for i in range(scan.np):
            p.pack_farray(scan.npts, scan.p[i].data[j][k], p.pack_double)
        for i in range(scan.nd):
            p.pack_farray(scan.npts, scan.d[i].data[j][k], p.pack_float)

    return(p.get_buffer())

def writeMDA(dim, fname=None):
    m = mdaBuf()
    p = xdr.Packer()

    p.reset()
    if (type(dim) != type([])): print("writeMDA: first arg must be a scan")
    if ((fname != None) and (type(fname) != type(""))):
        print("writeMDA: second arg must be a filename or None")
    rank = dim[0]['rank']    # rank of scan as a whole
    # write file header
    p.pack_float(dim[0]['version'])
    p.pack_int(dim[0]['scan_number'])
    p.pack_int(dim[0]['rank'])
    p.pack_farray(rank, dim[0]['dimensions'], p.pack_int)
    p.pack_int(dim[0]['isRegular'])
    m.header = p.get_buffer()

    p.reset()
    p.pack_int(0) # pExtra
    m.pExtra = p.get_buffer()

    m.scan = packScanHead(dim[1])
    m.scan.offset = len(m.header) + len(m.pExtra)
    m.scan.data = packScanData(dim[1], [])
    m.scan.bufLen = m.scan.bufLen + len(m.scan.data)
    prevScan = m.scan
    #print("\n m.scan=", m.scan)
    #print("\n type(m.scan.pLowerScans)=", type(m.scan.pLowerScans))

    if (rank > 1):
        for i in range(m.scan.npts):
            m.scan.inner.append(packScanHead(dim[2]))
            thisScan = m.scan.inner[i]
            thisScan.offset = prevScan.offset + prevScan.bufLen
            m.scan.pLowerScans.append(thisScan.offset)
            thisScan.data = packScanData(dim[2], [i])
            thisScan.bufLen = thisScan.bufLen + len(thisScan.data)
            prevScan = thisScan

            if (rank > 2):
                for j in range(m.scan.inner[i].npts):
                    m.scan.inner[i].inner.append(packScanHead(dim[3]))
                    thisScan = m.scan.inner[i].inner[j]
                    thisScan.offset = prevScan.offset + prevScan.bufLen
                    m.scan.inner[i].pLowerScans.append(thisScan.offset)
                    thisScan.data = packScanData(dim[3], [i,j])
                    thisScan.bufLen = thisScan.bufLen + len(thisScan.data)
                    prevScan = thisScan

                if (rank > 3):
                    for k in range(m.scan.inner[i].inner[j].npts):
                        m.scan.inner[i].inner[j].append(packScanHead(dim[4]))
                        thisScan = m.scan.inner[i].inner[j].inner[k]
                        thisScan.offset = prevScan.offset + prevScan.bufLen
                        m.scan.inner[i].inner[j].pLowerScans.append(thisScan.offset)
                        thisScan.data = packScanData(dim[4], [i,j,k])
                        thisScan.bufLen = thisScan.bufLen + len(thisScan.data)
                        prevScan = thisScan

    # Now we know where the extraPV section must go.
    p.reset()
    p.pack_int(prevScan.offset + prevScan.bufLen) # pExtra
    m.pExtra = p.get_buffer()

    # pack scan-environment variables from dictionary
    p.reset()

    numKeys = 0
    for name in dim[0].keys():
        if not (name in dim[0]['ourKeys']):
            numKeys = numKeys + 1
    p.pack_int(numKeys)

    for name in dim[0].keys():
        # Note we don't want to write the dict entries we made for our own
        # use in the scanDim object.
        if not (name in dim[0]['ourKeys']):
            desc = dim[0][name][0]
            unit = dim[0][name][1]
            value = dim[0][name][2]
            EPICS_type = dim[0][name][3]
            count = dim[0][name][4]
            n = len(name); p.pack_int(n)
            if (n): p.pack_string(name)
            n = len(desc); p.pack_int(n)
            if (n): p.pack_string(desc)
            p.pack_int(EPICS_type)
            if EPICS_type != 0:   # not DBR_STRING, so pack count and units
                p.pack_int(count)
                n = len(unit); p.pack_int(n)
                if (n): p.pack_string(unit)
            if EPICS_type == 0: # DBR_STRING
                n = len(value); p.pack_int(n)
                if (n): p.pack_string(value)
            elif EPICS_type == 32: # DBR_CTRL_CHAR
                # write null-terminated string
                v = []
                for i in range(len(value)): v.append(ord(value[i:i+1]))
                v.append(0)
                p.pack_farray(count, v, p.pack_int)
            elif EPICS_type == 29: # DBR_CTRL_SHORT
                p.pack_farray(count, value, p.pack_int)
            elif EPICS_type == 33: # DBR_CTRL_LONG
                p.pack_farray(count, value, p.pack_int)
            elif EPICS_type == 30: # DBR_CTRL_FLOAT
                p.pack_farray(count, value, p.pack_float)
            elif EPICS_type == 34: # DBR_CTRL_DOUBLE
                p.pack_farray(count, value, p.pack_double)

    m.extraPV = p.get_buffer()

    # Now we have to repack all the scan offsets
    if (rank > 1): # 2D scan
        #print("m.scan.pLowerScans", m.scan.pLowerScans)
        p.reset()
        p.pack_farray(m.scan.npts, m.scan.pLowerScans, p.pack_int)
        m.scan.pLowerScansBuf = p.get_buffer()
        if (rank > 2): # 3D scan
            for i in range(m.scan.npts):
                p.reset()
                p.pack_farray(m.scan.inner[i].npts, m.scan.inner[i].pLowerScans, p.pack_int)
                m.scan.inner[i].pLowerScansBuf = p.get_buffer()
                if (rank > 3): # 4D scan
                    for j in range(m.scan.inner[i].npts):
                        p.reset()
                        p.pack_farray(m.scan.inner[i].inner[j].npts, m.scan.inner[i].inner[j].pLowerScans, p.pack_int)
                        m.scan.inner[i].inner[j].pLowerScansBuf = p.get_buffer()

    # Write
    if (fname == None): fname = tkFileDialog.SaveAs().show()
    f = open(fname, 'wb')

    f.write(m.header)
    f.write(m.pExtra)
    s0 = m.scan
    f.write(s0.preamble)
    if len(s0.pLowerScansBuf): f.write(s0.pLowerScansBuf)
    f.write(s0.postamble)
    f.write(s0.data)
    for s1 in s0.inner:
        f.write(s1.preamble)
        if len(s1.pLowerScansBuf): f.write(s1.pLowerScansBuf)
        f.write(s1.postamble)
        f.write(s1.data)
        for s2 in s1.inner:
            f.write(s2.preamble)
            if len(s2.pLowerScansBuf): f.write(s2.pLowerScansBuf)
            f.write(s2.postamble)
            f.write(s2.data)
            for s3 in s2.inner:
                f.write(s3.preamble)
                if len(s3.pLowerScansBuf): f.write(s3.pLowerScansBuf)
                f.write(s3.postamble)
                f.write(s3.data)
    f.write(m.extraPV)
    f.close()
    return

################################################################################
# write Ascii file
def getFormat(d, rank):
    # number of positioners, detectors
    np = d[rank].np
    nd = d[rank].nd

    min_column_width = 15
    # make sure there's room for the names, etc.
    phead_fmt = []
    dhead_fmt = []
    pdata_fmt = []
    ddata_fmt = []
    columns = 1
    for i in range(np):
        cw = max(min_column_width, len(d[rank].p[i].name)+1)
        cw = max(cw, len(d[rank].p[i].desc)+1)
        cw = max(cw, len(d[rank].p[i].fieldName)+1)
        phead_fmt.append("%%-%2ds" % cw)
        pdata_fmt.append("%%- %2d.8f" % cw)
        columns = columns + cw
    for i in range(nd):
        cw = max(min_column_width, len(d[rank].d[i].name)+1)
        cw = max(cw, len(d[rank].d[i].desc)+1)
        cw = max(cw, len(d[rank].d[i].fieldName)+1)
        dhead_fmt.append("%%-%2ds" % cw)
        ddata_fmt.append("%%- %2d.8f" % cw)
        columns = columns + cw
    return (phead_fmt, dhead_fmt, pdata_fmt, ddata_fmt, columns)

def writeAscii(d, fname=None):
    if (type(d) != type([])):
        print("writeMDA: first arg must be a scan")
        return

    if (fname == None):
        f = sys.stdout
    else:
        f = open(fname, 'wb')

    (phead_fmt, dhead_fmt, pdata_fmt, ddata_fmt, columns) = getFormat(d, 1)
    # header
    f.write("### %s is a %d-dimensional file\n" % (d[0]['filename'], d[0]['rank']))
    f.write("### Number of data points      = [")
    for i in range(d[0]['rank'],1,-1): f.write("%-d," % d[i].curr_pt)
    f.write("%-d]\n" % d[1].curr_pt)

    f.write("### Number of detector signals = [")
    for i in range(d[0]['rank'],1,-1): f.write("%-d," % d[i].nd)
    f.write("%-d]\n" % d[1].nd)

    # scan-environment PV values
    f.write("#\n# Scan-environment PV values:\n")
    ourKeys = d[0]['ourKeys']
    maxKeyLen = 0
    for i in d[0].keys():
        if (i not in ourKeys):
            if len(i) > maxKeyLen: maxKeyLen = len(i)
    for i in d[0].keys():
        if (i not in ourKeys):
            f.write("#%s%s%s\n" % (i, (maxKeyLen-len(i))*' ', d[0][i]))

    f.write("\n#%s\n" % str(d[1]))
    f.write("#  scan date, time: %s\n" % d[1].time)
    sep = "#"*columns + "\n"
    f.write(sep)

    # 1D data table head
    f.write("#")
    for j in range(d[1].np):
        f.write(phead_fmt[j] % (d[1].p[j].fieldName))
    for j in range(d[1].nd):
        f.write(dhead_fmt[j] % (d[1].d[j].fieldName))
    f.write("\n")

    f.write("#")
    for j in range(d[1].np):
        f.write(phead_fmt[j] % (d[1].p[j].name))
    for j in range(d[1].nd):
        f.write(dhead_fmt[j] % (d[1].d[j].name))
    f.write("\n")

    f.write("#")
    for j in range(d[1].np):
        f.write(phead_fmt[j] % (d[1].p[j].desc))
    for j in range(d[1].nd):
        f.write(dhead_fmt[j] % (d[1].d[j].desc))
    f.write("\n")

    f.write(sep)

    # 1D data
    for i in range(d[1].curr_pt):
        f.write("")
        for j in range(d[1].np):
            f.write(pdata_fmt[j] % (d[1].p[j].data[i]))
        for j in range(d[1].nd):
            f.write(ddata_fmt[j] % (d[1].d[j].data[i]))
        f.write("\n")

    # 2D data
    if (len(d) > 2):
        f.write("\n# 2D data\n")
        for i in range(d[2].np):
            f.write("\n# Positioner %d (.%s) PV:'%s' desc:'%s'\n" % (i, d[2].p[i].fieldName, d[2].p[i].name, d[2].p[i].desc))
            for j in range(d[1].curr_pt):
                for k in range(d[2].curr_pt):
                    f.write("%f " % d[2].p[i].data[j][k])
                f.write("\n")

        for i in range(d[2].nd):
            f.write("\n# Detector %d (.%s) PV:'%s' desc:'%s'\n" % (i, d[2].d[i].fieldName, d[2].d[i].name, d[2].d[i].desc))
            for j in range(d[1].curr_pt):
                for k in range(d[2].curr_pt):
                    f.write("%f " % d[2].d[i].data[j][k])
                f.write("\n")

    if (len(d) > 3):
        f.write("\n# Can't write 3D (or higher) data\n")

    if (fname != None):
        f.close()


################################################################################
# misc
def showEnv(dict, all=0):
    if type(dict) == type([]) and type(dict[0]) == type({}):
        dict = dict[0]
    fieldLen = 0
    for k in dict.keys():
        if len(k) > fieldLen:
            fieldLen = len(k)
    format = "%%-%-ds %%s" % fieldLen
    for k in dict.keys():
        if not (k in dict['ourKeys']):
            if type(dict[k]) == type((1,2,3)):
                value = dict[k][2]
            else:
                value = dict[k]
            if type(value) == type([]) and len(value) == 1:
                value = value[0]
            if all:
                print(format % (k,dict[k]))
            else:
                print(format % (k,value))
    return

def fixMDA(d):
    """usage: d=fixMDA(d), where d is a list returned by readMDA()"""
    dimensions = []
    for i in range(1,len(d)):
        npts = d[i].curr_pt
        d[i].npts = npts
        dimensions.append(npts)
        for j in range(d[i].np):
            if (len(d[i].p[j].data) > npts):
                d[i].p[j].data = d[i].p[j].data[0:npts]
        for j in range(d[i].nd):
            if (len(d[i].d[j].data) > npts):
                d[i].d[j].data = d[i].d[j].data[0:npts]
    dimensions.reverse()
    d[0]['dimensions'] = dimensions
    return(d)

# translate mca-ROI PV's to mca-ROI description PV's, scaler signal PV'ss to scaler signal description PV's
descDict = {'R1':'R1NM', 'R2':'R2NM', 'R3':'R3NM', 'R4':'R4NM', 'R5':'R5NM',
 'R6':'R6NM', 'R7':'R7NM', 'R8':'R8NM', 'R9':'R9NM', 'R10':'R10NM',
 'R11':'R11NM', 'R12':'R12NM', 'R13':'R13NM', 'R14':'R14NM', 'R15':'R15NM',
 'R16':'R16NM', 'R17':'R17NM', 'R18':'R18NM', 'R19':'R19NM', 'R20':'R20NM',
 'R21':'R21NM', 'R22':'R22NM', 'R23':'R23NM', 'R24':'R24NM', 'R25':'R25NM',
 'R26':'R26NM', 'R27':'R27NM', 'R28':'R28NM', 'R29':'R29NM', 'R30':'R30NM',
 'R31':'R31NM', 'R32':'R32NM',
 'S1':'NM1', 'S2':'NM2', 'S3':'NM3', 'S4':'NM4', 'S5':'NM5', 'S6':'NM6', 'S7':'NM7', 'S8':'NM8', 'S9':'NM9', 'S10':'NM10',
 'S11':'NM11', 'S12':'NM12', 'S13':'NM13', 'S14':'NM14', 'S15':'NM15', 'S16':'NM16', 'S17':'NM17', 'S18':'NM18', 'S19':'NM19', 'S20':'NM20',
 'S21':'NM21', 'S22':'NM22', 'S23':'NM23', 'S24':'NM24', 'S25':'NM25', 'S26':'NM26', 'S27':'NM27', 'S28':'NM28', 'S29':'NM29', 'S30':'NM30',
 'S31':'NM31', 'S32':'NM32', 'S33':'NM33', 'S34':'NM34', 'S35':'NM35', 'S36':'NM36', 'S37':'NM37', 'S38':'NM38', 'S39':'NM39', 'S40':'NM40',
 'S41':'NM41', 'S42':'NM42', 'S43':'NM43', 'S44':'NM44', 'S45':'NM45', 'S46':'NM46', 'S47':'NM47', 'S48':'NM48', 'S49':'NM49', 'S50':'NM50',
 'S51':'NM51', 'S52':'NM52', 'S53':'NM53', 'S54':'NM54', 'S55':'NM55', 'S56':'NM56', 'S57':'NM57', 'S58':'NM58', 'S59':'NM59', 'S60':'NM60',
 'S61':'NM61', 'S62':'NM62', 'S63':'NM63', 'S64':'NM64'}

def findDescInEnv(name, env):
    try:
        (record, field) = name.split('.')
    except:
        return ""
    try:
        descField = descDict[field]
    except:
        return ""
    try:
        desc = env[record+'.'+descField]
    except:
        return ""
    if desc[2] == "" or desc[2].isspace():
        return ""
    return "{%s}" % desc[2]

def getDescFromEnv(data):
    if (data):
        for d in data[1:]:
            for p in d.p:
                if not p.desc:
                    p.desc = findDescInEnv(p.name, data[0])
            for d in d.d:
                if not d.desc:
                    d.desc = findDescInEnv(d.name, data[0])

########################
# opMDA and related code
########################
def isScan(d):
    if type(d) != type([]): return(0)
    if type(d[0]) != type({}): return(0)
    if 'rank' not in d[0].keys(): return(0)
    if len(d) < 2: return(0)
    if type(d[1]) != type(scanDim()): return(0)
    return(1)

def isScalar(d):
    if (type(d) == type(1)) or (type(d) == type(1.0)): return(1)
    return(0)

def add(a,b): return(a+b)
def sub(a,b): return(a-b)
def mul(a,b): return(a*b)
def div(a,b): return(a/b)

def setOp(op):
    if (op == '+') or (op == 'add'): return(add)
    if (op == '-') or (op == 'sub'): return(sub)
    if (op == '*') or (op == 'x') or (op == 'mul'): return(mul)
    if (op == '/') or (op == 'div'): return(div)
    if (op == '>') or (op == 'max'): return(max)
    if (op == '<') or (op == 'min'): return(min)
    print("opMDA: unrecognized op = ", op)
    return None

def opMDA_usage():
    print("opMDA() usage:")
    print("   result = opMDA(op, scan1, scan2)")
    print("        OR")
    print("   result = opMDA(op, scan1, scalar_value)")
    print("\nwhere:")
    print("   op is one of '+', '-', '*', '/', '>', '<'")
    print("   scan1, scan2 are scans, i.e., structures returned by mda.readMDA()")
    print("   result is a copy of scan1, modified by the operation\n")
    print("\n examples:")
    print("   r = opMDA('+', scan1, scan2) -- adds all detector data from scan1 and scan2")
    print("   r = opMDA('-', scan1, 2.0)   -- subtracts 2 from all detector data from scan1")
    print("   r = opMDA('>', r, 0)         -- 'r' data or 0, whichever is greater")

def opMDA_scalar(op, d1, scalar):
    op = setOp(op)
    if (op == None):
        opMDA_usage()
        return None

    s = copy.deepcopy(d1)

    # 1D op
    for i in range(s[1].nd):
        for j in range(s[1].npts):
            s[1].d[i].data[j] = op(s[1].d[i].data[j], scalar)

    if (len(s) == 2): return s
    # 2D op
    for i in range(s[2].nd):
        for j in range(s[1].npts):
            for k in range(s[2].npts):
                s[2].d[i].data[j][k] = op(s[2].d[i].data[j][k], scalar)

    if (len(s) == 3): return s
    # 3D op
    for i in range(s[3].nd):
        for j in range(s[1].npts):
            for k in range(s[2].npts):
                for l in range(s[3].npts):
                    s[3].d[i].data[j][k][l] = op(s[3].d[i].data[j][k][l], scalar)

    if (len(s) == 4): return s
    # 4D op
    for i in range(s[4].nd):
        for j in range(s[1].npts):
            for k in range(s[2].npts):
                for l in range(s[3].npts):
                    for m in range(s[4].npts):
                        s[4].d[i].data[j][k][l][m] = op(s[4].d[i].data[j][k][l][m], scalar)

    if (len(s) > 4):
        print("opMDA supports up to 4D scans")
    return s

def opMDA(op, d1, d2):
    """opMDA() is a function for performing arithmetic operations on MDA files,
    or on an MDA file and a scalar value.

    For examples, type 'opMDA_usage()'.
    """
    if isScan(d1) and isScalar(d2): return(opMDA_scalar(op,d1,d2))
    if (not isScan(d1)) :
        print("opMDA: first operand is not a scan")
        opMDA_usage()
        return None

    if (not isScan(d2)):
        print("opMDA: second operand is neither a scan nor a scalar")
        opMDA_usage()
        return None

    if len(d1) != len(d2):
        print("scans do not have same dimension")
        return None

    op = setOp(op)
    if (op == None):
        opMDA_usage()
        return None

    s = copy.deepcopy(d1)

    # 1D op
    if s[1].nd != d2[1].nd:
        print("scans do not have same number of 1D detectors")
        return None
    if s[1].npts != d2[1].npts:
        print("scans do not have same number of data points")
        return None
    for i in range(s[1].nd):
        s[1].d[i].data = map(op, s[1].d[i].data, d2[1].d[i].data)

    if (len(s) == 2): return s
    # 2D op
    if s[2].nd != d2[2].nd:
        print("scans do not have same number of 2D detectors")
        return None
    if s[2].npts != d2[2].npts:
        print("scans do not have same number of data points")
        return None
    for i in range(s[2].nd):
        for j in range(s[1].npts):
            s[2].d[i].data[j] = map(op, s[2].d[i].data[j], d2[2].d[i].data[j])

    if (len(s) == 3): return s
    # 3D op
    if s[3].nd != d2[3].nd:
        print("scans do not have same number of 3D detectors")
        return None
    if s[3].npts != d2[3].npts:
        print("scans do not have same number of data points")
        return None
    for i in range(s[3].nd):
        for j in range(s[1].npts):
            for k in range(s[2].npts):
                s[3].d[i].data[j][k] = map(op, s[3].d[i].data[j][k], d2[3].d[i].data[j][k])

    if (len(s) == 4): return s
    # 3D op
    if s[4].nd != d2[4].nd:
        print("scans do not have same number of 4D detectors")
        return None
    if s[4].npts != d2[4].npts:
        print("scans do not have same number of data points")
        return None
    for i in range(s[4].nd):
        for j in range(s[1].npts):
            for k in range(s[2].npts):
                for l in range(s[3].npts):
                    s[4].d[i].data[j][k][l] = map(op, s[4].d[i].data[j][k][l], d2[4].d[i].data[j][k][l])

    if (len(s) > 5):
        print("opMDA supports up to 4D scans")
    return s


def read_mda(fname, maxdim=3, verbose=False, _larch=None):
    """read an MDA file from the Epics Scan Record

    Arguments
    ---------
    filname (str)     name of file
    maxdim (integer)  max number of dimensions [default=3]

    Returns
    -------
    group containing `scan1` and possibly `scan2`, etc objects

    Notes
    -----
    not very well tested for scans of dimension > 2

    """
    out = readMDA(fname, maxdim=maxdim, verbose=verbose)
    if _larch is None:
        return out
    group = _larch.symtable.create_group(name='MDA_file %s' % fname)
    group.extra_pvs = out[0]
    group.scan1 = out[1]
    group.dimension = dim = len(out) - 1

    if dim > 1:
        group.scan2 = out[2]
    if dim > 2:
        group.scan3 = out[3]
    if dim > 3:
        group.scan4 = out[4]

    if dim == 1:
        data, array_labels, field_names, pv_names  = [], [], [], []
        for x in group.scan1.p + group.scan1.d:
            data.append(x.data)
            pv_names.append(x.name)
            field_names.append(x.fieldName)
            label = x.desc
            if label in (None, 'None', ''):
                label = x.fieldName
            array_labels.append(label)

        group.data = numpy.array(data)
        group.pv_names = pv_names
        group.field_names = field_names
        group.array_labels = array_labels
    return group



#!/usr/bin/env python
'''
SymbolTable for Larch interpreter
'''
from __future__ import print_function
import os
import sys
import types
import numpy
import copy
from . import site_config
from .closure import Closure
from .utils import fixName, isValidName


class Group(object):
    """
    Generic Group: a container for variables, modules, and subgroups.
    """
    __private = ('_main', '_larch', '_parents', '__name__', '__doc__',
                 '__private', '_subgroups', '_members', '_repr_html_')

    def __init__(self, name=None, **kws):
        if name is None:
            name = hex(id(self))
        self.__name__ = name
        for key, val in kws.items():
            setattr(self, key, val)

    def __len__(self):
        return max(1, len(dir(self))-1)

    def __repr__(self):
        if self.__name__ is not None:
            return '<Group %s>' % self.__name__
        return '<Group>'

    def __copy__(self):
        out = Group()
        for k, v in self.__dict__.items():
            if k != '__name__':
                setattr(out, k,  copy.copy(v))
        return out

    def __deepcopy__(self, memo):
        out = Group()
        for k, v in self.__dict__.items():
            if k != '__name__':
                setattr(out, k,  copy.deepcopy(v, memo))
        return out

    def __id__(self):
        return id(self)

    def __dir__(self):
        "return list of member names"
        cls_members = []
        cname = self.__class__.__name__
        if cname != 'SymbolTable' and hasattr(self, '__class__'):
            cls_members = dir(self.__class__)

        dict_keys = [key for key in self.__dict__ if key not in cls_members]

        return [key for key in cls_members + dict_keys
                if (not key.startswith('_SymbolTable_') and
                    not key.startswith('_Group_') and
                    not key.startswith('_%s_' % cname) and
                    not (key.startswith('__') and key.endswith('__')) and
                    key not in self.__private)]

    def _subgroups(self):
        "return list of names of members that are sub groups"
        return [k for k in self._members() if isgroup(self.__dict__[k])]

    def _members(self):
        "return members"
        r = {}
        for key in self.__dir__():
            if key in self.__dict__:
                r[key] = self.__dict__[key]
        return r

    def _repr_html_(self):
        """HTML representation for Jupyter notebook"""

        html = [f"Group {self.__name__}"]
        html.append("<table>")
        html.append("<tr><td><b>Attribute</b></td><td><b>Type</b></td></tr>")
        attrs = self.__dir__()
        atypes = [type(getattr(self, attr)).__name__ for attr in attrs]
        html.append(''.join([f"<tr><td>{attr}</td><td><i>{atp}</i></td></tr>" for attr, atp in zip(attrs, atypes)]))
        html.append("</table>")
        return ''.join(html)


def isgroup(grp, *args):
    """tests if input is a Group

    With additional arguments (all must be strings), it also tests
    that the group has an an attribute named for each argument. This
    can be used to test not only if a object is a Group, but whether
    it a group with expected arguments.
    """
    ret = isinstance(grp, Group)
    if ret and len(args) > 0:
        try:
            ret = all([hasattr(grp, a) for a in args])
        except TypeError:
            return False
    return ret


class InvalidName:
    """ used to create a value that will NEVER be a useful symbol.
    symboltable._lookup() uses this to check for invalid names"""
    pass

GroupDocs = {}
GroupDocs['_sys'] = """
Larch system-wide status variables, including
configuration variables and lists of Groups used
for finding variables.
"""

GroupDocs['_builtin'] = """
core built-in functions, most taken from Python
"""

GroupDocs['_math'] = """
Mathematical functions, including a host of functtion from numpy and scipy
"""


class SymbolTable(Group):
    """Main Symbol Table for Larch.
    """
    top_group   = '_main'
    core_groups = ('_sys', '_builtin', '_math')
    __invalid_name = InvalidName()
    _private = ('save_frame', 'restore_frame', 'set_frame',
                'has_symbol', 'has_group', 'get_group',
                'create_group', 'new_group', 'isgroup',
                'get_symbol', 'set_symbol',  'del_symbol',
                'get_parent', 'add_plugin', '_path', '__parents')

    def __init__(self, larch=None):
        Group.__init__(self, name=self.top_group)
        self._larch = larch
        self._sys = None
        setattr(self, self.top_group, self)

        for gname in self.core_groups:
            thisgroup = Group(name=gname)
            if gname in GroupDocs:
                thisgroup.__doc__ = GroupDocs[gname]

            setattr(self, gname, thisgroup)

        self._sys.frames      = []
        self._sys.searchGroups = [self.top_group]
        self._sys.path        = ['.']
        self._sys.localGroup  = self
        self._sys.valid_commands = []
        self._sys.moduleGroup = self
        self._sys.__cache__  = [None]*4
        self._sys.saverestore_groups = []
        for g in self.core_groups:
            self._sys.searchGroups.append(g)
        self._sys.core_groups = tuple(self._sys.searchGroups[:])

        self.__callbacks = {}
        orig_sys_path = sys.path[:]

        if site_config.modules_path is not None:
            for idir in site_config.modules_path:
                idirfull = os.path.abspath(idir)
                if idirfull not in self._sys.path and os.path.exists(idirfull):
                    self._sys.path.append(idirfull)

        sys.path = self._sys.path[:]
        for idir in orig_sys_path:
            idirfull = os.path.abspath(idir)
            if idirfull not in sys.path:
                sys.path.append(idirfull)

        self._sys.modules = {'_main':self}
        for gname in self.core_groups:
            self._sys.modules[gname] = getattr(self, gname)
        self._fix_searchGroups()

        self._sys.config = Group(home_dir    = site_config.home_dir,
                                 history_file= site_config.history_file,
                                 init_files  = site_config.init_files,
                                 modules_path= site_config.modules_path,
                                 plugins_path= site_config.plugins_path,
                                 user_larchdir= site_config.usr_larchdir,
                                 larch_version= site_config.larch_version)

    def save_frame(self):
        " save current local/module group"
        self._sys.frames.append((self._sys.localGroup,
                                 self._sys.moduleGroup))

    def restore_frame(self):
        "restore last saved local/module group"
        try:
            lgrp, mgrp = self._sys.frames.pop()
            self._sys.localGroup = lgrp
            self._sys.moduleGroup  = mgrp
            self._fix_searchGroups()
        except:
            pass

    def set_frame(self, groups):
        "set current execution frame (localGroup, moduleGroup)"
        self._sys.localGroup, self._sys.moduleGroup  = groups
        self._fix_searchGroups()


    def _fix_searchGroups(self, force=False):
        """resolve list of groups to search for symbol names:

        The variable self._sys.searchGroups holds the list of group
        names for searching for symbol names.  A user can set this
        dynamically.  The names need to be absolute (that is, relative to
        _main, and can omit the _main prefix).

        This calclutes and returns self._sys.searchGroupObjects,
        which is the list of actual group objects (not names) resolved from
        the list of names in _sys.searchGroups)

        _sys.localGroup,_sys.moduleGroup come first in the search list,
        followed by any search path associated with that module (from
        imports for that module)
        """
        ##
        # check (and cache) whether searchGroups needs to be changed.
        sys = self._sys
        cache = sys.__cache__
        if len(cache) < 4:
            cache = [None]*4
        if (sys.localGroup   == cache[0] and
            sys.moduleGroup  == cache[1] and
            sys.searchGroups == cache[2] and
            cache[3] is not None and not force):
            return cache[3]

        if sys.moduleGroup is None:
            sys.moduleGroup = self.top_group
        if sys.localGroup is None:
            sys.localGroup = self.moduleGroup

        cache[0] = sys.localGroup
        cache[1] = sys.moduleGroup
        snames  = []
        sgroups = []
        for grp in (sys.localGroup, sys.moduleGroup):
            if grp is not None and grp not in sgroups:
                sgroups.append(grp)
                snames.append(grp.__name__)

        sysmods = list(self._sys.modules.values())
        searchGroups  = sys.searchGroups[:]
        searchGroups.extend(self._sys.core_groups)
        for name in searchGroups:
            grp = None
            if name in self._sys.modules:
                grp = self._sys.modules[name]
            elif hasattr(self, name):
                gtest = getattr(self, name)
                if isinstance(gtest, Group):
                    grp = gtest
            elif '.' in name:
                parent, child= name.split('.')
                for sgrp in sysmods:
                    if (parent == sgrp.__name__ and
                        hasattr(sgrp, child)):
                        grp = getattr(sgrp, child)
                        break
            else:
                for sgrp in sysmods:
                    if hasattr(sgrp, name):
                        grp = getattr(sgrp, name)
                        break
            if grp is not None and grp not in sgroups:
                sgroups.append(grp)
                snames.append(name)

        self._sys.searchGroups = cache[2] = snames[:]
        sys.searchGroupObjects = cache[3] = sgroups[:]
        return sys.searchGroupObjects

    def get_parentpath(self, sym):
        """ get parent path for a symbol"""
        obj = self._lookup(sym)
        if obj is None:
            return
        out = []
        for s in reversed(self.__parents):
            if s.__name__ != '_main' or '_main' not in out:
                out.append(s.__name__)
        out.reverse()
        return '.'.join(out)

    def _lookup(self, name=None, create=False):
        """looks up symbol in search path
        returns symbol given symbol name,
        creating symbol if needed (and create=True)"""
        debug = False # not ('force'in name)
        if debug:  print( '====\nLOOKUP ', name)
        searchGroups = self._fix_searchGroups()
        self.__parents = []
        if self not in searchGroups:
            searchGroups.append(self)

        def public_attr(grp, name):
            return (hasattr(grp, name)  and
                    not (grp is self and name in self._private))

        parts = name.split('.')
        if len(parts) == 1:
            for grp in searchGroups:
                if public_attr(grp, name):
                    self.__parents.append(grp)
                    return getattr(grp, name)

        # more complex case: not immediately found in Local or Module Group
        parts.reverse()
        top   = parts.pop()
        out   = self.__invalid_name
        if top == self.top_group:
            out = self
        else:
            for grp in searchGroups:
                if public_attr(grp, top):
                    self.__parents.append(grp)
                    out = getattr(grp, top)
        if out is self.__invalid_name:
            raise NameError("'%s' is not defined" % name)

        if len(parts) == 0:
            return out

        while parts:
            prt = parts.pop()
            if hasattr(out, prt):
                out = getattr(out, prt)
            elif create:
                val = None
                if len(parts) > 0:
                    val = Group(name=prt)
                setattr(out, prt, val)
                out = getattr(out, prt)
            else:
                raise LookupError(
                    "cannot locate member '%s' of '%s'" % (prt,out))
        return out

    def has_symbol(self, symname):
        try:
            g = self.get_symbol(symname)
            return True
        except (LookupError, NameError, ValueError):
            return False

    def has_group(self, gname):
        try:
            g = self.get_group(gname)
            return True
        except (NameError, LookupError):
            return False

    def isgroup(self, sym):
        "test if symbol is a group"
        return isgroup(sym)

    def get_group(self, gname):
        "find group by name"
        sym = self._lookup(gname, create=False)
        if isgroup(sym):
            return sym
        else:
            raise LookupError(
                "symbol '%s' found, but not a group" % (gname))

    def create_group(self, **kw):
        "create a new Group, not placed anywhere in symbol table"
        return Group(**kw)

    def new_group(self, name, **kws):
        name = fixName(name)
        grp = Group(__name__ = name, **kws)
        self.set_symbol(name, value=grp)
        return grp

    def get_symbol(self, sym, create=False):
        "lookup and return a symbol by name"
        return self._lookup(sym, create=create)

    def set_symbol(self, name, value=None, group=None):
        "set a symbol in the table"
        grp = self._sys.localGroup
        if group is not None:
            grp = self.get_group(group)
        names = []

        for n in name.split('.'):
            if not isValidName(n):
                raise SyntaxError("invalid symbol name '%s'" % n)
            names.append(n)

        child = names.pop()
        for nam in names:
            if hasattr(grp, nam):
                grp = getattr(grp, nam)
                if not isgroup(grp):
                    raise ValueError(
                "cannot create subgroup of non-group '%s'" % grp)
            else:
                setattr(grp, nam, Group())

        setattr(grp, child, value)
        if (grp, child) in self.__callbacks:
            for func, args, kws in self.__callbacks[(grp, child)]:
                kws.update({'group': grp, 'value': value,
                            'symbolname': child})
                func(*args, **kws)
        return getattr(grp, child)

    def del_symbol(self, name):
        "delete a symbol"
        sym = self._lookup(name, create=False)
        parent, child = self.get_parent(name)
        self.clear_callbacks(name)
        delattr(parent, child)

    def clear_callbacks(self, name, index=None):
        """clear 1 or all callbacks for a symbol
        """
        parent, child = self.get_parent(name)
        if child is not None and (parent, child) in self.__callbacks:
            if index is not None and index <= len(self.__callbacks[(parent, child)]):
                self.__callbacks[(parent, child)].pop(index)
            else:
                while self.__callbacks[(parent, child)]:
                    self.__callbacks[(parent, child)].pop()

    def add_callback(self, name, func, args=None, kws=None):
        """set a callback to be called when set_symbol() is called
        for a named variable
        """
        try:
            var = self.get_symbol(name)
        except NameError:
            raise NameError(
                "cannot locate symbol '%s' for callback" % (name))
        key = self.get_parent(name)
        if key not in self.__callbacks:
            self.__callbacks[key] = []
        if args is None: args = ()
        if kws is None: kws = {}

        self.__callbacks[key].append((func, args, kws))

    def get_parent(self, name):
        """return parent group, child name for an absolute symbol name
        (as from _lookup) that is, a pair suitable for hasattr,
        getattr, or delattr
        """
        tnam = name.split('.')
        if len(tnam) < 1 or name == self.top_group:
            return (self, None)
        child = tnam.pop()
        sym = self
        if len(tnam) > 0:
            sym = self._lookup('.'.join(tnam))
        return sym, child

    def add_plugin(self, plugin, on_error, **kws):
        """Add a plugin: a module that includes a
        registerLarchPlugin function that returns
        larch_group_name, dict_of_symbol/functions
        """
        if not isinstance(plugin, types.ModuleType):
            on_error("%s is not a valid larch plugin" % repr(plugin))

        group_registrar = getattr(plugin, 'registerLarchGroups', None)
        if callable(group_registrar):
            savegroups = group_registrar()
            for group in savegroups:
                self._sys.saverestore_groups.append(group)

        registrar = getattr(plugin, 'registerLarchPlugin', None)
        if registrar is None:
            return

        groupname, syms = registrar()
        if not isinstance(syms, dict):
            raise ValueError('add_plugin requires dictionary of plugins')

        if not self.has_group(groupname):
            self.new_group(groupname)

        if groupname not in self._sys.searchGroups:
            self._sys.searchGroups.append(groupname)
        self._fix_searchGroups(force=True)

        for key, val in syms.items():
            if hasattr(val, '__call__') and hasattr(val, '__code__'): # is a function
                # test whether plugin func has a '_larch' kw arg
                #    __code__.co_flags & 8 == 'uses **kws'
                kws.update({'func': val, '_name':key})
                try:
                    nvars = val.__code__.co_argcount
                    if ((val.__code__.co_flags &8 != 0) or
                        '_larch' in val.__code__.co_varnames[:nvars]):
                        kws.update({'_larch':  self._larch})
                    val = Closure(**kws)
                except AttributeError: # cannot make a closure
                    pass
            self.set_symbol("%s.%s" % (groupname, key), val)

        plugin_init = getattr(plugin, 'initializeLarchPlugin', None)
        if plugin_init is not None:
            plugin_init(_larch=self._larch)
        return (groupname, syms)

    def show_group(self, groupname):
        """display group members --- simple version for tests"""
        out = []
        try:
            group = self.get_group(groupname)
        except (NameError, LookupError):
            return 'Group %s not found' % groupname

        title = group.__name__
        members = dir(group)
        out = ['== %s: %i symbols ==' % (title, len(members))]
        for item in members:
            obj = getattr(group, item)
            dval = None
            if isinstance(obj, numpy.ndarray):
                if len(obj) > 10 or len(obj.shape)>1:
                    dval = "array<shape=%s, type=%s>" % (repr(obj.shape),
                                                         repr(obj.dtype))
            if dval is None:
                dval = repr(obj)
            out.append('  %s: %s' % (item, dval))
        self._larch.writer.write("%s\n" % '\n'.join(out))



import numpy as np
import ctypes
from ctypes import POINTER, pointer, c_int, c_long, c_char, c_char_p, c_double

import larch
from larch.larchlib import get_dll
from xraydb import atomic_mass

F8LIB = None

FEFF_maxpts = 150  # nex
FEFF_maxpot = 11   # nphx
FEFF_maxleg = 9    # legtot
BOHR = 0.52917721067
RYDBERG = 13.605698

def with_phase_file(fcn):
    """decorator to ensure that the wrapped function either
    has a non-None 'phase_file' argument or that that
    self.phase_file is not None
    """
    errmsg = "function '%s' needs a non-None phase_file"
    def wrapper(*args, **keywords):
        "needs phase_file"
        phase_file = keywords.get('phase_file', None)
        if phase_file is None:
            phase_file = getattr(args[0], 'phase_file', None)
            if phase_file is None:
                raise AttributeError(errmsg % fcn.__name__)
        else:
            setattr(args[0], 'phase_file', phase_file)
        # raise Warning(errmsg % fcn.__name__)
        return fcn(*args, **keywords)
    wrapper.__doc__ = fcn.__doc__
    wrapper.__name__ = fcn.__name__
    wrapper.__filename__ = fcn.__code__.co_filename
    wrapper.__dict__.update(fcn.__dict__)
    return wrapper

class Feff8L_XAFSPath(object):
    """Feff8 Scattering Path calculation

    A calculation requires a Potentials and Phase Shift calculation
    in PAD format from Feff8L, and a list of scattering paths

    Usage:
    ------
       # create path
       path  = Feff8L_XAFSPath(phase_file='phase.pad')

       # list 'ipot' and labels for absorber, scatterers
       path.list_scatterers()

       # set coords for absorbing atom
       path.set_absorber(x=0., y=0., z=0.)

       # add scattering atom
       path.add_scatterer(x=1.5, y=1.5, z=1.5, ipot=1)

       # calculate basic (unaltered) XAFS contributions
       path.calcuate_xafs()

    """
    def __init__(self, phase_file=None, title=''):
        global F8LIB
        if F8LIB is None:
            try:
                F8LIB = get_dll('feff8lpath')
            except:
                pass
        self.reset(phase_file=phase_file, title=title)

    def reset(self, phase_file=None, title=''):
        """reset all path data"""
        self.phase_file = None
        if phase_file is not None:
            self.phase_file = phase_file
        self.index   = 9999
        self.degen   = 1.
        self.nnnn_out = False
        self.json_out = False
        self.verbose  = False
        self.ipol   = 0
        self.ellip  = 0.
        self.nepts  = 0
        self.genfmt_order = 2
        self.version= ""
        self.exch   = ""
        self.title  = title
        self.filename  = "%s_%s" % (self.phase_file, self.title)
        self.rs_int = 0.
        self.vint   = 0.
        self.mu     = 0.
        self.edge   = 0.
        self.kf     = 0.
        self.rnorman = 0.
        self.gam_ch = 0.
        self.nepts  = FEFF_maxpts

        dargs = dict(dtype=np.float64, order='F')
        largs = dict(dtype=np.int32, order='F')

        self.evec   = np.zeros(3, **dargs)
        self.xivec  = np.zeros(3, **dargs)
        self.ipot   = np.zeros(1+FEFF_maxleg, **largs)
        self.beta   = np.zeros(1+FEFF_maxleg, **dargs)
        self.eta    = np.zeros(2+FEFF_maxleg, **dargs)
        self.ri     = np.zeros(FEFF_maxleg, **dargs)
        self.rat    = np.zeros((3, 2+FEFF_maxleg), **dargs)
        self.iz     = np.zeros(1+FEFF_maxpot, **largs)
        self.k      = np.zeros(FEFF_maxpts, **dargs)
        self.real_phc = np.zeros(FEFF_maxpts, **dargs)
        self.mag_feff = np.zeros(FEFF_maxpts, **dargs)
        self.pha_feff = np.zeros(FEFF_maxpts, **dargs)
        self.red_fact = np.zeros(FEFF_maxpts, **dargs)
        self.lam      = np.zeros(FEFF_maxpts, **dargs)
        self.rep      = np.zeros(FEFF_maxpts, **dargs)
        self.nleg = 1
        self.atoms = []

        if self.phase_file is not None:
            self.read_atoms()
            self.set_absorber()

    @with_phase_file
    def read_atoms(self):
        """read atoms ipot, iz, symbol"""
        self.atoms = []
        with open(self.phase_file,'r') as fh:
            line1_words = fh.readline().strip().split()
            text = fh.readlines()
        npots = int(line1_words[4])
        for line in text[4:]:
            if not line.startswith('$'):
                words = line.split()
                self.atoms.append((int(words[1]), words[2]))
            if len(self.atoms) > npots:
                break

    @with_phase_file
    def list_atoms(self):
        """list Feff Potentials atoms ('ipots') fo phase file"""
        if len(self.atoms) < 1:
            self.read_atoms()
        out = ["# Potential   Z   Symbol"]
        for ipot, atom in enumerate(self.atoms):
            out.append("    %2i      %3i     %s" % (ipot, atom[0], atom[1]))
        return "\n".join(out)

    @with_phase_file
    def set_absorber(self, x=0., y=0., z=0., phase_file=None):
        """set coordinates for absorbing atom ('ipot'=0)"""
        self.rat[0, 0] = x
        self.rat[1, 0] = y
        self.rat[2, 0] = z
        self.rat[0, self.nleg] = self.rat[0, 0]
        self.rat[1, self.nleg] = self.rat[1, 0]
        self.rat[2, self.nleg] = self.rat[2, 0]
        self.ipot[self.nleg]   = self.ipot[0]

    @with_phase_file
    def add_scatterer(self, x=0., y=0., z=0., ipot=1, phase_file=None):
        self.rat[0, self.nleg] = x
        self.rat[1, self.nleg] = y
        self.rat[2, self.nleg] = z
        self.ipot[self.nleg] = ipot
        self.nleg += 1
        # set final atom coords to same as absorber
        self.rat[0, self.nleg] = self.rat[0, 0]
        self.rat[1, self.nleg] = self.rat[1, 0]
        self.rat[2, self.nleg] = self.rat[2, 0]
        self.ipot[self.nleg]   = self.ipot[0]

    @with_phase_file
    def calculate_xafs(self, phase_file=None):
        if F8LIB is None:
            raise ValueError("Feff8 Dynamic library not found")

        if len(self.atoms) < 1:
            self.read_atoms()

        class args:
            pass

        # strings / char*.  Note fixed length to match Fortran
        args.phase_file     = (self.phase_file + ' '*256)[:256]
        args.exch_label     = ' '*8
        args.genfmt_version = ' '*30

        # integers, including booleans
        for attr in ('index', 'nleg', 'genfmt_order', 'ipol', 'nnnn_out',
                     'json_out', 'verbose', 'nepts'):
            setattr(args, attr, pointer(c_long(int(getattr(self, attr)))))

        # doubles
        for attr in ('degen', 'rs_int', 'vint', 'mu', 'edge', 'kf', 'rnorman',
                     'gam_ch', 'ellip'):
            setattr(args, attr, pointer(c_double(getattr(self, attr))))

        # integer arrays
        args.ipot = self.ipot.ctypes.data_as(POINTER(self.ipot.size*c_int))
        args.iz = self.iz.ctypes.data_as(POINTER(self.iz.size*c_int))

        # double arrays
        # print(" Rat 0  ", self.rat)
        for attr in ('evec', 'xivec', 'ri', 'beta', 'eta',
                     'k', 'real_phc', 'mag_feff', 'pha_feff',
                     'red_fact', 'lam', 'rep'):
            arr = getattr(self, attr)
            cdata = arr.ctypes.data_as(POINTER(arr.size*c_double))
            setattr(args, attr, cdata)
        # handle rat (in atomic units)
        rat_atomic = self.rat/BOHR
        args.rat = (rat_atomic).ctypes.data_as(POINTER(rat_atomic.size*c_double))

        onepath = F8LIB.calc_onepath
        # print(" Calc with onepath ", onepath, rat_atomic)
        # print(" args rat = ", args.rat.contents[:])
        x = onepath(args.phase_file, args.index, args.nleg, args.degen,
                    args.genfmt_order, args.exch_label, args.rs_int, args.vint,
                    args.mu, args.edge, args.kf, args.rnorman,
                    args.gam_ch, args.genfmt_version, args.ipot, args.rat,
                    args.iz, args.ipol, args.evec, args.ellip, args.xivec,
                    args.nnnn_out, args.json_out, args.verbose, args.ri,
                    args.beta, args.eta, args.nepts, args.k,
                    args.real_phc, args.mag_feff, args.pha_feff,
                    args.red_fact, args.lam, args.rep)

        self.exch   = args.exch_label.strip()
        self.version = args.genfmt_version.strip()
        # print(" Calc with onepath done")
        # unpack integers/floats
        for attr in ('index', 'nleg', 'genfmt_order', 'degen', 'rs_int',
                     'vint', 'mu', 'edge', 'kf', 'rnorman', 'gam_ch',
                     'ipol', 'ellip', 'nnnn_out', 'json_out', 'verbose',
                     'nepts'):
            setattr(self, attr, getattr(args, attr).contents.value)

        # some data needs recasting, reformatting
        self.mu *= (2*RYDBERG)
        self.nnnn_out = bool(self.nnnn_out)
        self.json_out = bool(self.json_out)
        self.verbose  = bool(self.verbose)

        # unpck energies
        for attr in ('evec', 'xivec'):
            cdata = getattr(args, attr).contents[:]
            setattr(self, attr, np.array(cdata))

        nleg = self.nleg
        nepts = self.nepts

        # arrays of length 'nepts'
        for attr in ('k', 'real_phc', 'mag_feff', 'pha_feff',
                     'red_fact', 'lam', 'rep'):
            cdata = getattr(args, attr).contents[:nepts]
            setattr(self, attr, np.array(cdata))
        self.pha = self.real_phc + self.pha_feff
        self.amp = self.red_fact * self.mag_feff

        # unpack arrays of length 'nleg':
        for attr in ('ipot', 'beta', 'eta', 'ri'):
            cdata = getattr(args, attr).contents[:]
            setattr(self, attr, np.array(cdata))

        # rat is sort of special, and calculate reff too:
        rat = args.rat.contents[:]
        rat = np.array(rat).reshape(2+FEFF_maxleg, 3).transpose()
        self.rat = BOHR*rat

        _rat = self.rat.T
        reff = 0.
        for i, atom in enumerate(_rat[1:]):
            prev = _rat[i,:]
            reff += np.sqrt( (prev[0]-atom[0])**2 +
                             (prev[1]-atom[1])**2 +
                             (prev[2]-atom[2])**2 )
        self.reff = reff /2.0

        self.geom = []
        rmass  = 0.
        for i in range(nleg):
            ipot = int(self.ipot[i])
            iz, sym = self.atoms[ipot]
            mass = atomic_mass(iz)

            x, y, z = _rat[i][0], _rat[i][1], _rat[i][2]
            self.geom.append((str(sym), iz, ipot, x, y, z))
            rmass += 1.0/max(1.0, mass)

        self.rmass = 1./rmass


def feff8_xafs(phase_file):
    return Feff8L_XAFSPath(phase_file=phase_file)



## def initializeLarchPlugin(_larch=None):
#     """initialize F8LIB"""
#     if _larch is not None:
#         global F8LIB
#         if F8LIB is None:
#             try:
#                 F8LIB = get_dll('feff8lpath')
#             except:
#                 pass
##



#!/usr/bin/env pythonw
'''
popup for 2D XRD calibration

'''
import os
import numpy as np

import wx

from wxmplot.imagepanel import ImagePanel
from larch.io import tifffile
from larch.xrd import lambda_from_E, E_from_lambda

from .ImageControlsFrame import ImageToolboxFrame


HAS_pyFAI = False
try:
    import pyFAI
    import pyFAI.calibrant
    #from pyFAI.calibration import Calibration
    HAS_pyFAI = True
except ImportError:
    pass


###################################

class CalibrationPopup(wx.Frame):

    def __init__(self,parent):

        self.frame = wx.Frame.__init__(self, parent, title='Calibration',size=(900,700))

        self.parent = parent

        self.statusbar = self.CreateStatusBar(2,wx.CAPTION )
        self.default_cal = 0
        self.default_det = 0
        self.img_fname = ''

        try:
            self.raw_img = parent.plt_img ## raw_img or flp_img or plt_img mkak 2016.10.28
            self.img_fname = 'Image from diFFit2D viewer.'
        except:
            self.raw_img = np.zeros((1024,1024))

        self.Init()
        self.Show()

#        wx.Window.GetEffectiveMinSize
#        wx.GetBestSize(self)

        self.setDefaults()



    def Init(self):

        self.panel = wx.Panel(self)

        self.DirectionsSizer()
        self.MainSizer()
#        self.OKsizer()

        self.framebox = wx.BoxSizer(wx.VERTICAL)
        self.framebox.Add(self.dirbox,  flag=wx.ALL|wx.EXPAND, border=10)
        self.framebox.Add(self.mainbox, flag=wx.ALL|wx.EXPAND, border=10)
#        self.framebox.Add(self.okbox,   flag=wx.ALL|wx.ALIGN_RIGHT, border=10)

        ###########################
        ## Pack all together in self.panel
        self.panel.SetSizer(self.framebox)

        ###########################
        ## Set default information
        self.stepno = 0
        self.checkRANGE()
        self.showDirection()

    def setDefaults(self):

        ## Sets some typical defaults specific to GSE 13-ID procedure
        self.entr_pix.SetValue('400')     ## binned pixels (2x200um)
        self.entr_EorL.SetValue('19.0')     ## 19.0 keV
        self.entr_dist.SetValue('0.5')  ## 0.5 m
        self.ch_det.SetSelection(self.default_det)  ## Perkin detector
        self.ch_cal.SetSelection(self.default_cal)  ## CeO2
        self.entr_calimg.SetValue(self.img_fname)

        self.entr_cntrx.SetValue(str(int(self.raw_img.shape[0]/2))) ## x-position of beam
        self.entr_cntry.SetValue(str(int(self.raw_img.shape[1]/2))) ## y-position of beam

        self.onDorPSel()

    def DirectionsSizer(self):

        ###########################
        ## Directions
        dirbx = wx.StaticBox(self.panel,label='DIRECTIONS', size=(100, 50))
        self.dirbox = wx.StaticBoxSizer(dirbx,wx.VERTICAL)

        hbox_direct = wx.BoxSizer(wx.HORIZONTAL)
        self.followdir = wx.StaticText(self.panel,label='')

        #hbox_direct.Add(self.txt_shp, flag=wx.RIGHT, border=8)
        hbox_direct.Add(self.followdir, flag=wx.ALL|wx.EXPAND, border=8)

        self.dirbox.Add(hbox_direct, flag=wx.ALL|wx.EXPAND, border=10)

        hbox_next = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_prev = wx.Button(self.panel,label='PREVIOUS')
        self.btn_next = wx.Button(self.panel,label='NEXT')

        self.btn_prev.Bind(wx.EVT_BUTTON,self.onPREVIOUS)
        self.btn_next.Bind(wx.EVT_BUTTON,self.onNEXT)

        hbox_next.Add(self.btn_prev, flag=wx.ALL, border=8)
        hbox_next.Add((-1, 100))
        hbox_next.Add(self.btn_next, flag=wx.ALIGN_RIGHT|wx.ALL, border=8)

        self.dirbox.Add(hbox_next, flag=wx.ALL|wx.EXPAND, border=10)

    def MainSizer(self):

        self.mainbox = wx.BoxSizer(wx.VERTICAL)

        ###########################
        ## -----> Main Panel
        self.hmain = wx.BoxSizer(wx.HORIZONTAL)

        self.ImageSizer()
        self.ParameterSizer()

        self.hmain.Add(self.imagebox,proportion=1,flag=wx.ALL|wx.EXPAND, border=10)
        self.hmain.Add(self.parbox, flag=wx.ALL, border=10)

        self.mainbox.Add(self.hmain, flag=wx.ALL|wx.EXPAND, border=10)

    def ParameterSizer(self):
        '''
        This is where the parameters will be.
        '''

        #self.parbox = wx.BoxSizer(wx.VERTICAL)
        prbx = wx.StaticBox(self.panel,label='PARAMETERS', size=(50, 100))
        self.parbox = wx.StaticBoxSizer(prbx,wx.VERTICAL)

        ###########################
        ## Establish lists from pyFAI
        clbrnts = [] #['None']
        self.dets = [] #['None']
        for key,value in pyFAI.detectors.ALL_DETECTORS.items():
            self.dets.append(key)
            if key == 'perkin':
                self.default_det = len(self.dets)-1
        for key,value in pyFAI.calibrant.ALL_CALIBRANTS.items():
            clbrnts.append(key)
            if key == 'CeO2':
                self.default_cal = len(clbrnts)-1


        #####
        ## Calibration Image selection
        hbox_cal1  = wx.BoxSizer(wx.HORIZONTAL)
        ttl_calimg = wx.StaticText(self.panel, label='Calibration Image:' )
        self.entr_calimg = wx.TextCtrl(self.panel, size=(210, -1))
#         btn_calimg = wx.Button(self.panel, label='Browse...')

#         btn_calimg.Bind(wx.EVT_BUTTON, self.loadIMAGE)

        hbox_cal1.Add(ttl_calimg,       flag=wx.RIGHT,  border=8)
        hbox_cal1.Add(self.entr_calimg, flag=wx.RIGHT|wx.EXPAND,  border=8)
#         hbox_cal1.Add(btn_calimg,       flag=wx.RIGHT,  border=8)
        self.parbox.Add(hbox_cal1,      flag=wx.BOTTOM|wx.TOP, border=8)

        btn_calimg = wx.Button(self.panel, label='Browse...')
        btn_calimg.Bind(wx.EVT_BUTTON, self.loadIMAGE)
        self.parbox.Add(btn_calimg,      flag=wx.BOTTOM|wx.ALIGN_RIGHT, border=8)

        #####
        ## Calibrant selection
        hbox_cal2 = wx.BoxSizer(wx.HORIZONTAL)
        ttl_cal = wx.StaticText(self.panel, label='Calibrant:')
        self.ch_cal = wx.Choice(self.panel,choices=clbrnts)

        self.ch_cal.Bind(wx.EVT_CHOICE,  self.onCalSel)

        hbox_cal2.Add(ttl_cal,     flag=wx.RIGHT,  border=8)
        hbox_cal2.Add(self.ch_cal, flag=wx.RIGHT,  border=8)
        self.parbox.Add(hbox_cal2, flag=wx.BOTTOM, border=30)

        #####
        ## Set-up specific parameters
        hbox_cal3 = wx.BoxSizer(wx.HORIZONTAL)
        txt_exp = wx.StaticText(self.panel, label='SET-UP PARAMETERS')
        btn_pni = wx.Button(self.panel, label='Load file')

        btn_pni.Bind(wx.EVT_BUTTON,  self.openPONI)

        hbox_cal3.Add(txt_exp,     flag=wx.RIGHT,  border=8)
        hbox_cal3.Add(btn_pni,     flag=wx.LEFT,  border=60)
        self.parbox.Add(hbox_cal3, flag=wx.BOTTOM, border=8)

        #####
        ## Detector selection
        hbox_cal4 = wx.BoxSizer(wx.HORIZONTAL)
        self.ch_DorP = wx.Choice(self.panel,choices=['Detector name','Pixel size (um)'])
        self.ch_det  = wx.Choice(self.panel, choices=self.dets)
        self.entr_pix    = wx.TextCtrl(self.panel, size=(110, -1))

        self.ch_det.Bind(wx.EVT_CHOICE,  self.onDetSel)
        self.ch_DorP.Bind(wx.EVT_CHOICE, self.onDorPSel)

        hbox_cal4.Add(self.ch_DorP,  flag=wx.RIGHT,  border=8)
        hbox_cal4.Add(self.ch_det,   flag=wx.RIGHT,  border=8)
        hbox_cal4.Add(self.entr_pix, flag=wx.RIGHT,  border=8)
        self.parbox.Add(hbox_cal4,   flag=wx.BOTTOM, border=8)

        #####
        ## Energy or Wavelength
        hbox_cal5 = wx.BoxSizer(wx.HORIZONTAL)
        self.ch_EorL = wx.Choice(self.panel,choices=['Energy (keV)','Wavelength (A)'])
        self.entr_EorL = wx.TextCtrl(self.panel, size=(110, -1))

        self.ch_EorL.Bind(wx.EVT_CHOICE, self.onEorLSel)

        hbox_cal5.Add(self.ch_EorL,   flag=wx.RIGHT,  border=8)
        hbox_cal5.Add(self.entr_EorL, flag=wx.RIGHT,  border=8)
        self.parbox.Add(hbox_cal5,    flag=wx.BOTTOM, border=8)

        ## Distance
        hbox_cal6 = wx.BoxSizer(wx.HORIZONTAL)
        ttl_dist = wx.StaticText(self.panel, label='Detector distance (m):')
        self.entr_dist = wx.TextCtrl(self.panel, size=(110, -1))

        hbox_cal6.Add(ttl_dist,       flag=wx.RIGHT,  border=8)
        hbox_cal6.Add(self.entr_dist, flag=wx.RIGHT,  border=8)

        self.parbox.Add(hbox_cal6,   flag=wx.BOTTOM, border=8)

        ## Beam center x
        hbox_cal7 = wx.BoxSizer(wx.HORIZONTAL)
        ttl_cntrx = wx.StaticText(self.panel, label='Beam center, x (pixels):')
        self.entr_cntrx = wx.TextCtrl(self.panel, size=(110, -1))

        hbox_cal7.Add(ttl_cntrx,       flag=wx.RIGHT,  border=8)
        hbox_cal7.Add(self.entr_cntrx, flag=wx.RIGHT,  border=8)

        self.parbox.Add(hbox_cal7,   flag=wx.BOTTOM, border=8)

        ## Beam center y
        hbox_cal8 = wx.BoxSizer(wx.HORIZONTAL)
        ttl_cntry = wx.StaticText(self.panel, label='Beam center, y (pixels):')
        self.entr_cntry = wx.TextCtrl(self.panel, size=(110, -1))

        hbox_cal8.Add(ttl_cntry,       flag=wx.RIGHT,  border=8)
        hbox_cal8.Add(self.entr_cntry, flag=wx.RIGHT,  border=8)

        self.parbox.Add(hbox_cal8,   flag=wx.BOTTOM, border=8)


    def onCalSel(self,event=None):
        print('Selected calibrant: %s' % self.ch_cal.GetString(self.ch_cal.GetSelection()))

    def onDetSel(self,event=None):
        print('Selected detector: %s' % self.ch_det.GetString(self.ch_det.GetSelection()))


    def onEorLSel(self,event=None):

        if self.ch_EorL.GetSelection() == 1:
            energy = float(self.entr_EorL.GetValue()) ## units keV
            wavelength = lambda_from_E(energy) ## units: A
            self.entr_EorL.SetValue(str(wavelength))
        else:
            wavelength = float(self.entr_EorL.GetValue())*1e-10 ## units: m
            energy = E_from_lambda(wavelength) ## units: keV
            self.entr_EorL.SetValue(str(energy))

    def onDorPSel(self,event=None):
        if self.ch_DorP.GetSelection() == 0:
            self.entr_pix.Hide()
            self.ch_det.Show()
        else:
            self.ch_det.Hide()
            self.entr_pix.Show()

        self.panel.GetSizer().Layout()
        self.panel.GetParent().Layout()

    def loadIMAGE(self,event=None):
        wildcards = 'XRD image (*.edf,*.tif,*.tiff)|*.tif;*.tiff;*.edf|All files (*.*)|*.*'
        if os.path.exists(self.entr_calimg.GetValue()):
           dfltDIR = self.entr_calimg.GetValue()
        else:
           dfltDIR = os.getcwd()

        dlg = wx.FileDialog(self, message='Choose XRD calibration file',
                           defaultDir=dfltDIR,
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            try:
#                 self.raw_img = plt.imread(path)
                self.raw_img = tifffile.imread(path)
                #self.raw_img = fabio.open(path).data
            except:
                print('Image not properly opened.')
                pass
            self.plot2Dimg.display(self.raw_img)
            self.plot2Dimg.redraw()
            self.AutoContrast()

            self.entr_calimg.Clear()
            self.entr_calimg.SetValue(path) #os.path.split(path)[-1]


    def ImageSizer(self):
        '''
        Image Panel
        '''
        self.imagebox = wx.BoxSizer(wx.VERTICAL)

        self.plot2Dimage()

        self.btn_image = wx.Button(self.panel,label='IMAGE TOOLS')

        self.btn_image.Bind(wx.EVT_BUTTON,self.onImageTools)

        self.imagebox.Add(self.plot2Dimg,proportion=1,flag=wx.ALL|wx.EXPAND, border=10)
        self.imagebox.Add(self.btn_image, flag=wx.ALL, border=10)

#     def OKsizer(self):
#         ###########################
#         ## OK - CANCEL
#         self.okbox = wx.BoxSizer(wx.HORIZONTAL)
#
#         okBtn  = wx.Button(self.panel, wx.ID_OK     )
#         canBtn = wx.Button(self.panel, wx.ID_CANCEL )
#
#         self.okbox.Add(canBtn,  flag=wx.RIGHT, border=5)
#         self.okbox.Add(okBtn,   flag=wx.RIGHT, border=5)

    def write_message(self, s, panel=0):
        """write a message to the Status Bar"""
        self.SetStatusText(s, panel)

    def onImageTools(self,event=None):

        self.toolbox = ImageToolboxFrame(self.plot2Dimg,self.raw_img)

    def plot2Dimage(self):

        self.plot2Dimg = ImagePanel(self.panel,size=(300, 300))
        self.plot2Dimg.messenger = self.write_message

        self.plot2Dimg.display(self.raw_img)
        self.AutoContrast()

        self.plot2Dimg.redraw()

    def AutoContrast(self):

        self.minINT = int(np.min(self.raw_img))
        self.maxINT = int(np.max(self.raw_img)/15) # /15 scales image to viewable
        if self.maxINT == self.minINT:
            self.minINT = self.minINT-50
            self.maxINT = self.minINT+100

        self.minCURRENT = self.minINT
        self.maxCURRENT = self.maxINT
        if self.maxCURRENT > self.maxINT:
            self.maxCURRENT = self.maxINT

        self.plot2Dimg.conf.auto_intensity = False
        self.plot2Dimg.conf.int_lo[0] = self.minCURRENT
        self.plot2Dimg.conf.int_hi[0] = self.maxCURRENT
#         self.plot2Dimg.conf.int_lo['int'] = self.minCURRENT
#         self.plot2Dimg.conf.int_hi['int'] = self.maxCURRENT

        ## vertical flip default
        self.plot2Dimg.conf.flip_ud = True
        self.plot2Dimg.conf.flip_lr = False

        self.plot2Dimg.redraw()

    def checkRANGE(self):

        if self.stepno <= 0:
            self.stepno = 0
            self.btn_prev.Disable()
        else:
            self.btn_prev.Enable()

        if self.stepno >= 8:
            self.stepno = 8
            self.btn_next.Disable()
        else:
            self.btn_next.Enable()

    def onNEXT(self,event=None):
        self.stepno = self.stepno + 1
        self.checkRANGE()
        self.showDirection()

    def onPREVIOUS(self,event=None):
        self.stepno = self.stepno - 1
        self.checkRANGE()
        self.showDirection()

    def showDirection(self):

        dirsteps = ['Enter parameters into the fields below.',
                    'Select point(s) on the first ring.',
                    'Select point(s) on the second ring.',
                    'Select point(s) on the third ring.',
                    'Select point(s) on the fourth ring.',
                    'Select point(s) on the fifth ring.',
                    'Select point(s) on the sixth ring.',
                    'Check preliminary calibration. Continue for final refinement.',
                    'Refinement complete.' ]

        self.followdir.SetLabel(dirsteps[self.stepno])

    def openPONI(self,event=None):

        wildcards = 'pyFAI calibration file (*.poni)|*.poni|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose pyFAI calibration file',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:

            try:
                print
                self.ai = pyFAI.load(path)
                print('Loading calibration file: %s' % path)
            except:
                print('Not recognized as a pyFAI calibration file: %s' % path)
                return

            ## Sets viewer to values in .poni file
            self.entr_dist.SetValue('%0.4f' % self.ai._dist)
            self.entr_pix.SetValue('%0.1f' % float(self.ai.detector.pixel1*1000000.))
            self.ch_DorP.SetSelection(1)
            self.entr_EorL.SetValue('%0.4f' % float(self.ai._wavelength*1.e10))
            self.ch_EorL.SetSelection(1)
            self.onDorPSel()

            cenx = float(self.ai._poni1)/float(self.ai.detector.pixel1)
            ceny = float(self.ai._poni2)/float(self.ai.detector.pixel2)
            self.entr_cntrx.SetValue('%0.3f' % cenx)
            self.entr_cntry.SetValue('%0.3f' % ceny)

class CalXRD(wx.Dialog):
    """"""

    #----------------------------------------------------------------------
    def __init__(self):

        if HAS_pyFAI:
            ## Constructor
            dialog = wx.Dialog.__init__(self, None, title='XRD Calibration',size=(460, 440))
            ## remember: size=(width,height)
            self.panel = wx.Panel(self)

            self.InitUI()
            self.Centre()
            self.Show()

            ## Sets some typical defaults specific to GSE 13-ID procedure
            self.pixel.SetValue('400')     ## binned pixels (2x200um)
            self.EorL.SetValue('19.0')     ## 19.0 keV
            self.Distance.SetValue('0.5')  ## 0.5 m
            self.detslct.SetSelection(22)  ## Perkin detector
            self.calslct.SetSelection(20)  ## CeO2

            if self.slctDorP.GetSelection() == 0:
                self.sizer.Hide(self.pixel)

            ## Do not need flags if defaults are set
            #self.FlagCalibrant = False
            #self.FlagDetector  = False
            self.FlagCalibrant = True
            self.FlagDetector  = True

        else:
            print('pyFAI must be available for calibration.')
            return

    def InitUI(self):


        ## Establish lists from pyFAI
        clbrnts = [] #['None']
        self.dets = [] #['None']
        for key,value in pyFAI.detectors.ALL_DETECTORS.items():
            self.dets.append(key)
        for key,value in pyFAI.calibrant.ALL_CALIBRANTS.items():
            clbrnts.append(key)
        self.CaliPath = None


        ## Calibration Image selection
        caliImg     = wx.StaticText(self.panel,  label='Calibration Image:' )
        self.calFil = wx.TextCtrl(self.panel, size=(190, -1))
        fileBtn1    = wx.Button(self.panel,      label='Browse...'             )

        ## Calibrant selection
        self.calslct = wx.Choice(self.panel,choices=clbrnts)
        CalLbl = wx.StaticText(self.panel, label='Calibrant:' ,style=LEFT)

        ## Detector selection
        self.slctDorP = wx.Choice(self.panel,choices=['Detector','Pixel size (um)'])
        self.detslct  = wx.Choice(self.panel, choices=self.dets)
        self.pixel    = wx.TextCtrl(self.panel, size=(140, -1))

        ## Energy or Wavelength
        self.slctEorL = wx.Choice(self.panel,choices=['Energy (keV)','Wavelength (A)'])
        self.EorL = wx.TextCtrl(self.panel, size=(140, -1))

        ## Refine label
        RefLbl = wx.StaticText(self.panel, label='To be refined...' ,style=LEFT)

        ## Distance
        self.Distance = wx.TextCtrl(self.panel, size=(140, -1))
        DstLbl = wx.StaticText(self.panel, label='Distance (m):' ,style=LEFT)

        hlpBtn = wx.Button(self.panel, wx.ID_HELP   )
        okBtn  = wx.Button(self.panel, wx.ID_OK     )
        canBtn = wx.Button(self.panel, wx.ID_CANCEL )

        self.Bind(wx.EVT_BUTTON,   self.onBROWSE1,  fileBtn1 )
        self.calslct.Bind(wx.EVT_CHOICE,  self.onCalSel)
        self.detslct.Bind(wx.EVT_CHOICE,  self.onDetSel)
        self.slctDorP.Bind(wx.EVT_CHOICE, self.onDorPSel)
        self.slctEorL.Bind(wx.EVT_CHOICE, self.onEorLSel)

        self.sizer = wx.GridBagSizer(3, 3)

        self.sizer.Add(caliImg,       pos = ( 1,1)               )
        self.sizer.Add(self.calFil,   pos = ( 1,2), span = (1,2) )
        self.sizer.Add(fileBtn1,      pos = ( 1,4)               )
        self.sizer.Add(CalLbl,        pos = ( 3,1)               )
        self.sizer.Add(self.calslct,  pos = ( 3,2), span = (1,2) )

        self.sizer.Add(self.slctDorP, pos = ( 4,1)               )
        self.sizer.Add(self.detslct,  pos = ( 4,2), span = (1,4) )
        self.sizer.Add(self.pixel,    pos = ( 5,2), span = (1,2) )

        self.sizer.Add(self.slctEorL, pos = ( 6,1)               )
        self.sizer.Add(self.EorL,     pos = ( 6,2), span = (1,2) )

        self.sizer.Add(RefLbl,        pos = ( 8,1)               )
        self.sizer.Add(DstLbl,        pos = ( 9,1)               )
        self.sizer.Add(self.Distance, pos = ( 9,2), span = (1,2) )

        self.sizer.Add(hlpBtn,        pos = (11,1)               )
        self.sizer.Add(canBtn,        pos = (11,2)               )
        self.sizer.Add(okBtn,         pos = (11,3)               )

        self.FindWindowById(wx.ID_OK).Disable()

        self.panel.SetSizer(self.sizer)


    def onCalSel(self,event=None):
        #if self.calslct.GetSelection() == 0:
        #    self.FlagCalibrant = False
        #else:
        #    self.FlagCalibrant = True
        self.checkOK()

    def onDetSel(self,event=None):
        #if self.detslct.GetSelection() == 0:
        #    self.FlagDetector = False
        #else:
        #    self.FlagDetector = True
        self.checkOK()

    def onCheckOK(self,event=None):
        self.checkOK()

    def checkOK(self):
        if self.FlagCalibrant and self.CaliPath is not None:
            if self.slctDorP.GetSelection() == 1:
                self.FindWindowById(wx.ID_OK).Enable()
            else:
                if self.FlagDetector:
                    self.FindWindowById(wx.ID_OK).Enable()
                else:
                    self.FindWindowById(wx.ID_OK).Disable()
        else:
            self.FindWindowById(wx.ID_OK).Disable()

    def onEorLSel(self,event=None):

        if self.slctEorL.GetSelection() == 1:
            energy = float(self.EorL.GetValue()) ## units keV
            wavelength = lambda_from_E(energy) ## units: A
            self.EorL.SetValue(str(wavelength))
        else:
            wavelength = float(self.EorL.GetValue()) ## units: A
            energy = E_from_lambda(wavelength) ## units: keV
            self.EorL.SetValue(str(energy))

        self.checkOK()

    def onDorPSel(self,event=None):
        if self.slctDorP.GetSelection() == 0:
            self.sizer.Hide(self.pixel)
            self.sizer.Show(self.detslct)
        else:
            self.sizer.Hide(self.detslct)
            self.sizer.Show(self.pixel)

        self.checkOK()

    def onBROWSE1(self,event=None):
        wildcards = 'XRD image (*.edf,*.tif,*.tiff)|*.tif;*.tiff;*.edf|All files (*.*)|*.*'
        if os.path.exists(self.calFil.GetValue()):
           dfltDIR = self.calFil.GetValue()
        else:
           dfltDIR = os.getcwd()

        dlg = wx.FileDialog(self, message='Choose XRD calibration file',
                           defaultDir=dfltDIR,
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            self.calFil.Clear()
            self.calFil.SetValue(os.path.split(path)[-1])
            self.CaliPath = path
            self.checkOK()

# # # # # # # WAS IN mapviewer.py ; needs to be corrected or removed
#
# # # # def onCalXRD(self, evt=None):
# # # #     """
# # # #     Perform calibration with pyFAI
# # # #     mkak 2016.09.16
# # # #     """
# # # #
# # # #     ### can this pop up pyFAI or Dioptas GUI instead of creating own?
# # # #
# # # #     myDlg = CalXRD()
# # # #
# # # #     path, read = None, False
# # # #     if myDlg.ShowModal() == wx.ID_OK:
# # # #         read = True
# # # #
# # # #     myDlg.Destroy()
# # # #
# # # #     if read:
# # # #
# # # #         usr_calimg = myDlg.CaliPath
# # # #
# # # #         if myDlg.slctEorL.GetSelection() == 1:
# # # #             usr_lambda = float(myDlg.EorL.GetValue())*1e-10 ## units: m
# # # #             usr_E = E_from_lambda(usr_lambda,lambda_units='m') ## units: keV
# # # #         else:
# # # #             usr_E = float(myDlg.EorL.GetValue()) ## units keV
# # # #             usr_lambda = lambda_from_E(usr_E,lambda_units='m') ## units: m
# # # #
# # # #         if myDlg.slctDorP.GetSelection() == 1:
# # # #             usr_pixel = float(myDlg.pixel.GetValue())*1e-6
# # # #         else:
# # # #             usr_det  = myDlg.detslct.GetString(myDlg.detslct.GetSelection())
# # # #         usr_clbrnt  = myDlg.calslct.GetString(myDlg.calslct.GetSelection())
# # # #         usr_dist = float(myDlg.Distance.GetValue())
# # # #
# # # #         verbose = True #False
# # # #         if verbose:
# # # #             print('\n=== Calibration input ===')
# # # #             print('XRD image: %s' % usr_calimg)
# # # #             print('Calibrant: %s' % usr_clbrnt)
# # # #             if myDlg.slctDorP.GetSelection() == 1:
# # # #                 print('Pixel size: %0.1f um' % (usr_pixel*1e6))
# # # #             else:
# # # #                 print('Detector: %s' % usr_det)
# # # #             print('Incident energy: %0.2f keV (%0.4f A)' % (usr_E,usr_lambda*1e10))
# # # #             print('Starting distance: %0.3f m' % usr_dist)
# # # #             print('=========================\n')
# # # #
# # # #         ## Adapted from pyFAI-calib
# # # #         ## note: -l:units mm; -dist:units m
# # # #         ## mkak 2016.09.19
# # # #
# # # #         if myDlg.slctDorP.GetSelection() == 1:
# # # #             pform1 = 'pyFAI-calib -c %s -p %s -e %0.1f -dist %0.3f %s'
# # # #             command1 = pform1 % (usr_clbrnt,usr_pixel,usr_E,usr_dist,usr_calimg)
# # # #         else:
# # # #             pform1 = 'pyFAI-calib -c %s -D %s -e %0.1f -dist %0.3f %s'
# # # #             command1 = pform1 % (usr_clbrnt,usr_det,usr_E,usr_dist,usr_calimg)
# # # #         pform2 = 'pyFAI-recalib -i %s -c %s %s'
# # # #         command2 = pform2 % (usr_calimg.split('.')[0]+'.poni',usr_clbrnt,usr_calimg)
# # # #
# # # #         if verbose:
# # # #             print('\nNot functioning within code yet... but you could execute:')
# # # #             print('\t $ %s' % command1)
# # # #             print('\t $ %s\n\n' % command2)




# class diFFit_XRDcal(wx.App):
#     def __init__(self):
#         wx.App.__init__(self)
#
#     def run(self):
#         self.MainLoop()
#
#     def createApp(self):
#         frame = CalibrationPopup()
#         frame.Show()
#         self.SetTopWindow(frame)
#
#     def OnInit(self):
#         self.createApp()
#         return True
#

# class DebugViewer(diFFit_XRDcal):
#     def __init__(self, **kws):
#         diFFit_XRDcal.__init__(self, **kws)
#
#     def OnInit(self):
#         #self.Init()
#         self.createApp()
#         #self.ShowInspectionTool()
#         return True
#
# if __name__ == '__main__':
#     diFFit_XRDcal().run()



#!/usr/bin/env python
# coding: utf-8
"""This module provides integration of an IPython kernel.

.. note:: Initial idea taken from ipykernel example `internal_ipkernel.py`.
"""
from __future__ import absolute_import, division
import sys
from ipykernel import connect_qtconsole
from ipykernel.kernelapp import IPKernelApp


class InternalIPyKernel(object):

    def init_kernel(self, backend='qt', log_level='INFO'):
        _optslist = ['python',
                     '--gui={0}'.format(backend),
                     '--log-level={0}'.format(log_level)]

        self.kernel = IPKernelApp.instance()
        self.kernel.initialize(_optslist)

        # To create and track active qt consoles
        self.consoles = []

        # This application will also act on the shell user namespace
        self.namespace = self.kernel.shell.user_ns
        self.add_to_namespace('kernel', self.kernel)

    def print_namespace(self, evt=None):
        print("\n***Variables in User namespace***")
        for k, v in self.namespace.items():
            if not k.startswith('_'):
                print('%s -> %r' % (k, v))
        sys.stdout.flush()

    def add_to_namespace(self, namestr, nameobj):
        """Extend kernel namespace."""
        self.namespace[namestr] = nameobj

    def new_qt_console(self):
        """Start a new qtconsole connected to our kernel."""
        self.consoles.append(connect_qtconsole(self.kernel.abs_connection_file, profile=self.kernel.profile))

    def cleanup_consoles(self):
        for c in self.consoles:
            c.kill()



#!/usr/bin/env python
"""
General-Purpose Parameter widget
  shows Float TextCtrl for value, Choice fixed/variable/constrained
  and icon button to bring up dialog for full settings,
  including max/min value and constraint expression
"""

import numpy as np
import ast
import wx
from wx.lib.embeddedimage import PyEmbeddedImage

from wxutils import (GridPanel, Choice, FloatCtrl,
                     LEFT, pack, HLine, SetTip)

from lmfit import Parameter
from larch import Group
from larch.larchlib import Empty

PAR_FIX = 'fix'
PAR_VAR = 'vary'
PAR_CON = 'constrain'
VARY_CHOICES = (PAR_VAR, PAR_FIX, PAR_CON)


BOUNDS_custom = 'custom'
BOUNDS_none   = 'unbound'
BOUNDS_pos    = 'positive'
BOUNDS_neg    = 'negative'
BOUNDS_CHOICES = (BOUNDS_none, BOUNDS_pos, BOUNDS_neg, BOUNDS_custom)

PAR_WIDS = ('name', 'value',  'minval', 'maxval',
            'vary', 'expr', 'stderr', 'bounds')

infoicon = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAACzElEQVR42m2TW0gUURjHv3Nm"
    "b+6ut028lOaldm0RL2gGxWamSWHWi4JhPRjokyhCYWA9lfRQD0FUBJbdCBEEUzFILFjxQaxQ"
    "Wi+prXfXXV11J2zXXXdmOjPO7no78zAcvu//+/7nO+dDcMD6Ob14OiwisnDItpk0aXOi5XWX"
    "ecm68qW5Irtvby7aualq/q49k6FrlEulOQ4Pi4ALxLa8DFjsG/0zc9bKDzezhvcByt8PntRr"
    "Y7vlCmk4QmhXkGNZAIyBI0CXy7MxNW8vaizVG/05RQ9aNOmGHJNCIT3Mi3m9T3xcjSAvXgXG"
    "eSeM0SyBcOBxe1cnRn6ntVTlWoS86s6ZxwqV/DYWxXjHwc7HKcFwVA0dEzQMLbuBJS5YAnG7"
    "PK+eXo6vRNk36qisa1VWhHEEFsU+F5gkRisR8O1Y2eSE6tsAQmZZ5/i39ghUUN+UEpmZM8xX"
    "x0QtAMiXGEJBoS4MUmKCYWCOhre/1gMAlhNcOBfMBpR3911+cPKpHgoToQBAEKmkIDZUDqlR"
    "QXBBdwg6Ru3QOrImNFGA8ADSV2ZtsQQZ6prOKpIye4XqGPntKyQYGku0oJRJoLb9D8zT7j0A"
    "sqGtV1FaeUOUKq1gCRM1Eo/BX8FFXTjcyk+EcdsG1LRNCkIQ7YsQjlkwnRD6ndlgHGARzg44"
    "QPC8VA/pcWHw5OsU2Bwu6Jv+Kwh9DhDHTgzeO5csAPR1n4pZiboViQCVnALjHQNQ5PG0/ViA"
    "7jE79E85eP/bEAKgwFsx+vDSa/+NJ9R2fuQQVcY3QCbF8Kg0FWjXFrzpnYXZVacoBuGPgOly"
    "mrquLPe85PwATX61TKbNbeIAXxeeMto1JiCqSRO87cySqWyl475z3zDxS51bU0yFHqmnNAkZ"
    "/MX65Mz6rHlz9PMzxm5+4V2b2zpwGgMziiQoSBODQ6KPEa2EpS0WzuWwkMg/fjB3pv4HvQJH"
    "bUDKnS4AAAAASUVORK5CYII=")


class ParameterWidgets(object):
    """a set of related widgets for a lmfit Parameter

    param = Parameter(value=11.22, vary=True, min=0, name='x1')
    wid   = ParameterPanel(parent_wid, param)
    """
    def __init__(self, parent, param,  name_size=None, prefix=None,
                 expr_size=120, stderr_size=120, float_size=80,
                 minmax_size=60, widgets=PAR_WIDS):

        self.parent = parent
        self.param = param
        self._saved_expr = ''
        if (prefix is not None and
            not self.param.name.startswith(prefix)):
            self.param.name = "%s%s" %(prefix, self.param.name)

        for attr in PAR_WIDS:
            setattr(self, attr, None)

        # set vary_choice from param attributes
        vary_choice = PAR_VAR
        if param.expr not in (None, 'None', ''):
            vary_choice = PAR_CON
        elif not param.vary:
            vary_choice = PAR_FIX

        if 'name' in widgets:
            name = param.name
            if name in (None, 'None', ''):
                name = ''
            if name_size is None:
                name_size = min(50, len(param.name)*10)
            self.name = wx.StaticText(parent, label=name,
                                      size=(name_size, -1))

        if 'value' in widgets:
            self.value = FloatCtrl(parent, value=param.value,
                                   minval=param.min,
                                   maxval=param.max,
                                   action=self.onValue,
                                   act_on_losefocus=True,
                                   gformat=True,
                                   size=(float_size, -1))

        if 'minval' in widgets:
            minval = param.min
            if minval in (None, 'None', -np.inf):
                minval = -np.inf
            self.minval = FloatCtrl(parent, value=minval,
                                    gformat=True,
                                    size=(minmax_size, -1),
                                    act_on_losefocus=True,
                                    action=self.onMinval)
            self.minval.Enable(vary_choice==PAR_VAR)

        if 'maxval' in widgets:
            maxval = param.max
            if maxval in (None, 'None', np.inf):
                maxval = np.inf
            self.maxval = FloatCtrl(parent, value=maxval,
                                    gformat=True,
                                    size=(minmax_size, -1),
                                    act_on_losefocus=True,
                                    action=self.onMaxval)
            self.maxval.Enable(vary_choice==PAR_VAR)

        if 'vary' in widgets:
            self.vary = Choice(parent, size=(90, -1),
                               choices=VARY_CHOICES,
                               action=self.onVaryChoice)
            self.vary.SetStringSelection(vary_choice)

        if 'expr' in widgets:
            expr = param.expr
            if expr in (None, 'None', ''):
                expr = ''
            self._saved_expr = expr
            self.expr = wx.TextCtrl(parent, -1, value=expr,
                                      size=(expr_size, -1))
            self.expr.Enable(vary_choice==PAR_CON)
            self.expr.Bind(wx.EVT_CHAR, self.onExprChar)
            self.expr.Bind(wx.EVT_KILL_FOCUS, self.onExprKillFocus)
            SetTip(self.expr, 'Enter constraint expression')

        if 'stderr' in widgets:
            stderr = param.stderr
            if stderr in (None, 'None', ''):
                stderr = ''
            self.stderr = wx.StaticText(parent, label=stderr,
                                        size=(stderr_size, -1))

        if 'minval' in widgets or 'maxval' in widgets:
            minval = param.min
            maxval = param.max
            bounds_choice = BOUNDS_custom
            if minval in (None, 'None', -np.inf) and maxval in (None, 'None', np.inf):
                bounds_choice = BOUNDS_none
            elif minval == 0:
                bounds_choice = BOUNDS_pos
            elif maxval == 0:
                bounds_choice = BOUNDS_neg

            self.bounds = Choice(parent, size=(90, -1),
                                 choices=BOUNDS_CHOICES,
                                 action=self.onBOUNDSChoice)
            self.bounds.SetStringSelection(bounds_choice)

    def onBOUNDSChoice(self, evt=None):
        bounds = str(evt.GetString().lower())
        if bounds == BOUNDS_custom:
            pass
        elif bounds == BOUNDS_none:
            self.minval.SetValue(-np.inf)
            self.maxval.SetValue(np.inf)

        elif bounds == BOUNDS_pos:
            self.minval.SetValue(0)
            if float(self.maxval.GetValue()) == 0:
                self.maxval.SetValue(np.inf)
        elif bounds == BOUNDS_neg:
            self.maxval.SetValue(0)
            if float(self.minval.GetValue()) == 0:
                self.minval.SetValue(-np.inf)

    def onValue(self, evt=None, value=None):
        if value is not None:
            self.param.value = value

    def onExprChar(self, evt=None):
        key = evt.GetKeyCode()
        if key == wx.WXK_RETURN:
            self.onExpr(value=self.expr.GetValue())
        evt.Skip()

    def onExprKillFocus(self, evt=None):
        self.onExpr(value=self.expr.GetValue())
        evt.Skip()

    def onExpr(self, evt=None, value=None):
        if value is None and evt is not None:
            value = evt.GetString()
        try:
            ast.parse(value)
            self.param.expr = value
            bgcol, fgcol = 'white', 'black'
        except SyntaxError:
            bgcol, fgcol = 'red', 'yellow'

        self.expr.SetForegroundColour(fgcol)
        self.expr.SetBackgroundColour(bgcol)


    def onMinval(self, evt=None, value=None):
        if value in (None, 'None', ''):
            value = -np.inf
        if self.value is not None:
            v = self.value.GetValue()
            self.value.SetMin(value)
            self.value.SetValue(v)
            self.param.min = value
        if self.bounds is not None:
            if value == 0:
                self.bounds.SetStringSelection(BOUNDS_pos)
            elif value == -np.inf:
                if self.maxval.GetValue() == np.inf:
                    self.bounds.SetStringSelection(BOUNDS_none)
            else:
                self.bounds.SetStringSelection(BOUNDS_custom)

    def onMaxval(self, evt=None, value=None):
        # print "onMaxval " , value, self.value, self.value
        if value in (None, 'None', ''):
            value = np.inf
        if self.value is not None:
            v = self.value.GetValue()
            self.value.SetMax( value)
            self.value.SetValue(v)
            self.param.max = value

        if self.bounds is not None:
            if value == 0:
                self.bounds.SetStringSelection(BOUNDS_neg)
            elif value == np.inf:
                if self.minval.GetValue() == -np.inf:
                    self.bounds.SetStringSelection(BOUNDS_none)
            else:
                self.bounds.SetStringSelection(BOUNDS_custom)


    def onVaryChoice(self, evt=None):
        if self.vary is None:
            return
        vary = str(evt.GetString().lower())
        self.param.vary = (vary==PAR_VAR)
        if ((vary == PAR_VAR or vary == PAR_FIX) and
            self.param.expr not in (None, 'None', '')):
            self._saved_expr = self.param.expr
            self.param.expr = ''
        elif (vary == PAR_CON and self.param.expr in (None, 'None', '')):
            self.param.expr = self._saved_expr

        if self.value is not None:
            self.value.Enable(vary!=PAR_CON)
        if self.expr is not None:
            self.expr.Enable(vary==PAR_CON)
        if self.minval is not None:
            self.minval.Enable(vary==PAR_VAR)
        if self.maxval is not None:
            self.maxval.Enable(vary==PAR_VAR)
        if self.bounds is not None:
            self.bounds.Enable(vary==PAR_VAR)


class ParameterDialog(wx.Dialog):
    """Dialog (modal, that is, block) for Parameter Configuration"""
    def __init__(self, parent, param, precision=4, vary=None, **kws):
        self.param = param
        title = "  Parameter:  %s  " % (param.name)
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)
        panel = GridPanel(self)
        self.SetFont(parent.GetFont())

        if vary is None:
            vary = 0
            if param.vary:
                vary = 1
            elif param.expr is not None:
                vary = 2

        minval, maxval = param.min, param.max
        stderr, expr   = param.stderr, param.expr
        sminval = "%s" % minval
        smaxval = "%s" % maxval
        if minval in (None, 'None', -np.inf): minval = -np.inf
        if maxval in (None, 'None',  np.inf): maxval = np.inf
        if stderr is None: stderr = ''
        if expr is None:   expr = ''

        self.wids = Empty()
        self.wids.vary = Choice(panel, choices=VARY_CHOICES,
                                action=self.onVaryChoice, size=(110, -1))
        self.wids.vary.SetSelection(vary)

        self.wids.val  = FloatCtrl(panel, value=param.value, size=(100, -1),
                                   precision=precision,
                                   minval=minval, maxval=maxval)
        self.wids.min  = FloatCtrl(panel, value=minval, size=(100, -1))
        self.wids.max  = FloatCtrl(panel, value=maxval, size=(100, -1))
        self.wids.expr = wx.TextCtrl(panel, value=expr, size=(300, -1))
        self.wids.err  = wx.StaticText(panel, label="%s" % stderr)

        SetTip(self.wids.expr, "Mathematical expression to calcutate value")

        btnsizer = wx.StdDialogButtonSizer()
        ok_btn = wx.Button(panel, wx.ID_OK)
        ok_btn.SetDefault()
        btnsizer.AddButton(ok_btn)
        btnsizer.AddButton(wx.Button(panel, wx.ID_CANCEL))
        btnsizer.Realize()

        panel.AddText(' Name:',       style=LEFT)
        panel.AddText(param.name,     style=LEFT)
        panel.AddText(' Type:',       style=LEFT)
        panel.Add(self.wids.vary,     style=LEFT)
        panel.AddText(' Value:',      style=LEFT, newrow=True)
        panel.Add(self.wids.val,      style=LEFT)
        panel.AddText(' Std Error:',  style=LEFT)
        panel.Add(self.wids.err,      style=LEFT)
        panel.AddText(' Min Value:',  style=LEFT, newrow=True)
        panel.Add(self.wids.min,      style=LEFT)
        panel.AddText(' Max Value:',  style=LEFT)
        panel.Add(self.wids.max,      style=LEFT)
        panel.AddText(' Constraint:', style=LEFT, newrow=True)
        panel.Add(self.wids.expr,     style=LEFT, dcol=3)

        panel.Add(HLine(panel, size=(375, 2)), dcol=4, newrow=True)
        panel.Add(btnsizer,  dcol=4, newrow=True, style=LEFT)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 0, 0, 25)
        self.onVaryChoice()
        pack(self, sizer)
        bsize = self.GetBestSize()
        self.SetSize((bsize[0]+10, bsize[1]+10))

    def onVaryChoice(self, evt=None):
        if evt is not None:
            vary = evt.GetString()
        else:
            vary = self.wids.vary.GetStringSelection()
        if vary == PAR_CON:
            self.wids.val.Disable()
            self.wids.expr.Enable()
        else:
            self.wids.val.Enable()
            self.wids.expr.Disable()


class ParameterPanel(wx.Panel):
    """wx.Panel for a Larch Parameter

    param = Parameter(value=11.22, vary=True, min=0, name='x1')
    wid   = ParameterPanel(parent_wid, param)
    """
    def __init__(self, parent, param, size=(80, -1), show_name=False, precision=4, **kws):
        self.param = param
        self.precision = precision
        wx.Panel.__init__(self, parent, -1)
        self.wids = Empty()

        self.wids.val = FloatCtrl(self, value=param.value,
                                  minval=param.min, maxval=param.max,
                                  precision=precision, size=size)

        self.wids.name = None
        self.wids.edit = wx.Button(self, label='edit', size=(45, 25))
        self.wids.edit.Bind(wx.EVT_BUTTON, self.onConfigure)
        SetTip(self.wids.edit, "Configure Parameter")

        self.wids.vary = Choice(self, choices=VARY_CHOICES,
                                action=self.onVaryChoice, size=(80, -1))

        vary_choice = 0
        if param.vary:
            vary_choice = 1
        elif param.expr is not None:
            vary_choice = 2
        self.wids.vary.SetSelection(vary_choice)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        CLEFT = LEFT|wx.ALL
        if show_name:
            self.wids.name = wx.StaticText(self,
                                           label="%s: " %  param.name,
                                           size=(len(param.name)*8, -1))
            sizer.Add(self.wids.name,  0, CLEFT)

        sizer.Add(self.wids.val,   0, CLEFT)
        sizer.Add(self.wids.vary,  0, CLEFT)
        sizer.Add(self.wids.edit,  0, CLEFT)
        pack(self, sizer)

    def onVaryChoice(self, evt=None):
        vary = evt.GetString()
        self.param.vary = (vary == PAR_VAR)
        if vary == PAR_CON:
            self.wids.val.Disable()
        else:
            self.wids.val.Enable()

    def onConfigure(self, evt=None):
        self.param.value = self.wids.val.GetValue()
        vary = self.wids.vary.GetSelection()
        dlg = ParameterDialog(self, self.param, vary=vary,
                              precision=self.precision)
        dlg.Raise()
        if dlg.ShowModal() == wx.ID_OK:
            self.param.max   = float(dlg.wids.max.GetValue())
            self.param.min   = float(dlg.wids.min.GetValue())
            self.param.value = float(dlg.wids.val.GetValue())
            self.wids.val.SetMax(self.param.max)
            self.wids.val.SetMin(self.param.min)

            self.wids.val.SetValue(self.param.value)

            var = dlg.wids.vary.GetSelection()
            self.wids.vary.SetSelection(var)
            self.param.vary = False
            if var == 1:
                self.param.vary = True
            elif var == 2:
                self.param.expr= dlg.wids.expr.GetValue()
            self.param._getval()
        dlg.Destroy()

class TestFrame(wx.Frame):
    def __init__(self, parent=None, size=(-1, -1)):
        wx.Frame.__init__(self, parent, -1, 'Parameter Panel Test',
                          size=size, style=wx.DEFAULT_FRAME_STYLE)
        panel = GridPanel(self)

        param1 = Parameter(value=99.0, vary=True, min=0,            name='peak1_amplitude')
        param2 = Parameter(value=110.2, vary=True, min=100, max=120, name='peak1_center')
        param3 = Parameter(value=1.23, vary=True, min=0.5, max=2.0, name='peak1_sigma')

        panel.Add(ParameterPanel(panel, param1, show_name=True), style=LEFT)
        # panel.NewRow()
        panel.Add(ParameterPanel(panel, param2, show_name=True), style=LEFT)
        panel.Add(ParameterPanel(panel, param3, show_name=True), style=LEFT)
        panel.pack()
        self.createMenus()

        self.SetSize((700, 200))
        self.Show()
        self.Raise()

    def createMenus(self):
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        im = fmenu.Append(wid, "Show Widget Frame\tCtrl+I", "")
        sellf.Bind(wx.EVT_MENU, self.onShowInspection, im.Id)
        self.menubar.Append(fmenu, "&File")
        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE, self.onExit)

    def onShowInspection(self, evt=None):
        wx.GetApp().ShowInspectionTool()

    def onExit(self, evt=None):
        self.Destroy()

class TestApp(wx.App, wx.lib.mixins.inspection.InspectionMixin):
    def __init__(self, **kws):
        wx.App.__init__(self)

    def OnInit(self):
        self.Init()
        frame = TestFrame()
        frame.Show()
        self.SetTopWindow(frame)
        return True

if __name__ == "__main__":
    TestApp().MainLoop()



#!/usr/bin/env python
# coding: utf-8
"""
RIXS data view
"""
import os
from silx.gui import qt

from larch.qtlib.view import TreeView

from .rixsdata import RixsData
from .items import RixsItem


class RixsTreeView(TreeView):

    def __init__(self, parent=None):
        super(RixsTreeView, self).__init__(parent)

    def loadFiles(self):
        paths, _ = qt.QFileDialog.getOpenFileNames(
            self, 'Select Files to Load', os.getcwd(),
            'RixsData Files (*rixs.h5)')

        if not paths:
            return

        parent = self.selectionModel().selectedRows().pop()
        parentItem = self.model().itemFromIndex(parent)
        for path in paths:
            self.addFile(path, parentItem)

    def addFile(self, path=None, parentItem=None):
        if path is None:
            return

        # Add the file to the last added experiment item.
        if parentItem is None:
            parentItem = self.model().rootItem.lastChild()

        try:
            rdata = RixsData()
            rdata.load_from_h5(path)
        except Exception:
            return

        # Create a tree item for the file and add it to the experiment item.
        item = RixsItem(rdata.sample_name, parentItem, data=rdata)
        self.model().appendRow(item)

    def rixsItems(self):
        for index in self.model().visitModel():
            item = self.model().itemFromIndex(index)
            if isinstance(item, RixsItem):
                yield item


class RixsListView(qt.QListView):
    """Simple List View used in larch.qtrixs.plotrixs.RixsMainWindow"""

    def __init__(self, parent=None):
        super(RixsListView, self).__init__(parent)


if __name__ == '__main__':
    pass
